# -*- coding: utf-8 -*-

import os

import pytest
from mock import sentinel

from sceptre.helpers import get_subclasses
from sceptre.helpers import camel_to_snake_case
from sceptre.helpers import recurse_into_sub_environments
from sceptre.helpers import get_name_tuple
from sceptre.helpers import resolve_stack_name
from sceptre.helpers import get_external_stack_name
from sceptre.hooks import Hook
from sceptre.resolvers import Resolver


class TestHelpers(object):

    def test_get_subclasses(self):
        directory = os.path.join(os.getcwd(), "sceptre", "resolvers")
        classes = get_subclasses(Resolver, directory)

        # This is actually checking a property of the classes, which isn't
        # ideal but it's difficult to assert that the classes themselves are
        # the same.
        assert classes["environment_variable"].__name__ == \
            "EnvironmentVariable"
        assert classes["file_contents"].__name__ == \
            "FileContents"
        assert classes["stack_output_external"].__name__ == \
            "StackOutputExternal"
        assert classes["stack_output"].__name__ ==  \
            "StackOutput"
        assert classes["project_variables"].__name__ == \
            "ProjectVariables"
        assert len(classes) == 5

    def test_camel_to_snake_case(self):
        snake_case_string = camel_to_snake_case("Bash")
        assert snake_case_string == "bash"
        snake_case_string = camel_to_snake_case("ASGScheduledActions")
        assert snake_case_string == "asg_scheduled_actions"

    def test_recurse_into_sub_environments_with_leaf_object(self):
        class MockEnv(object):

            def __init__(self, name, is_leaf):
                self.name = name
                self.is_leaf = is_leaf

            @recurse_into_sub_environments
            def do(self):
                return {self.name: sentinel.response}

        mock_env = MockEnv("mock_stack", True)
        response = mock_env.do()
        assert response == {"mock_stack": sentinel.response}

    def test_recurse_into_sub_environments_with_non_leaf_object(self):
        class MockEnv(object):

            def __init__(self, name, is_leaf):
                self.name = name
                self.is_leaf = is_leaf

            @recurse_into_sub_environments
            def do(self):
                return {self.name: sentinel.response}

        mock_env = MockEnv("non-leaf-stack", False)

        # Add leaf sub-environments
        mock_env.environments = {
            "mock-env-1": MockEnv("leaf-stack-1", True),
            "mock-env-2": MockEnv("leaf-stack-2", True)
        }

        response = mock_env.do()
        assert response == {
            "leaf-stack-1": sentinel.response,
            "leaf-stack-2": sentinel.response
        }

    def test_get_name_tuple(self):
        result = get_name_tuple("dev/ew1/jump-host")
        assert result == ("dev", "ew1", "jump-host")

    def test_resolve_stack_name(self):
        result = resolve_stack_name("dev/ew1/subnet", "vpc")
        assert result == "dev/ew1/vpc"
        result = resolve_stack_name("dev/ew1/subnet", "prod/ue1/vpc")
        assert result == "prod/ue1/vpc"

    def test_get_external_stack_name(self):
        result = get_external_stack_name("prj", "dev/ew1/jump-host")
        assert result == "prj-dev-ew1-jump-host"

    def test_get_subclasses_with_invalid_directory(self):
        with pytest.raises(TypeError):
            get_subclasses(Hook, 1)



# -*- coding: utf-8 -*-

"""
sceptre.template

This module implements a Template class, which stores a CloudFormation template
and implements methods for uploading it to S3.
"""

from datetime import datetime
import imp
import logging
import os
import sys
import threading

import botocore
import jinja2
from .exceptions import UnsupportedTemplateFileTypeError
from .exceptions import TemplateSceptreHandlerError


class Template(object):
    """
    Template represents an AWS CloudFormation template. It is responsible for
    loading, storing, and optionally uploading, local templates for use by the
    CloudFormation service.

    :param path: The absolute path to the file which stores the template.
    :type path: str
    :param sceptre_user_data: A dictionary of arbitrary data to be passed to \
        a handler function in an external Python script.
    :type sceptre_user_data: dict
    """

    _boto_s3_lock = threading.Lock()

    def __init__(self, path, sceptre_user_data):
        self.logger = logging.getLogger(__name__)

        self.path = path
        self.sceptre_user_data = sceptre_user_data
        self.name = os.path.basename(path).split(".")[0]
        self._body = None

    def __repr__(self):
        return (
            "sceptre.template.Template(name='{0}', path='{1}', "
            "sceptre_user_data={2})".format(
                self.name, self.path, self.sceptre_user_data
            )
        )

    @property
    def body(self):
        """
        Represents body of the CloudFormation template.

        :returns: The body of the CloudFormation template.
        :rtype: str
        """
        if self._body is None:
            file_extension = os.path.splitext(self.path)[1]

            if file_extension in {".json", ".yaml"}:
                with open(self.path) as template_file:
                    self._body = template_file.read()
            elif file_extension == ".j2":
                self._body = self._render_jinja_template(
                    os.path.dirname(self.path),
                    os.path.basename(self.path),
                    {"sceptre_user_data": self.sceptre_user_data}
                )
            elif file_extension == ".py":
                self._body = self._call_sceptre_handler()

            else:
                raise UnsupportedTemplateFileTypeError(
                    "Template has file extension %s. Only .py, .yaml, "
                    ".json and .j2 are supported.",
                    os.path.splitext(self.path)[1]
                )
        return self._body

    def _call_sceptre_handler(self):
        """
        Calls the function `sceptre_handler` within templates that are python
        scripts.

        :returns: The string returned from sceptre_handler in the template.
        :rtype: str
        :raises: IOError
        :raises: TemplateSceptreHandlerError
        """
        # Get relative path as list between current working directory and where
        # the template is
        # NB: this is a horrible hack...
        relpath = os.path.relpath(self.path, os.getcwd()).split(os.path.sep)
        relpaths_to_add = [
            os.path.sep.join(relpath[:i+1])
            for i in range(len(relpath[:-1]))
        ]
        # Add any directory between the current working directory and where
        # the template is to the python path
        for directory in relpaths_to_add:
            sys.path.append(os.path.join(os.getcwd(), directory))
        self.logger.debug(
            "%s - Getting CloudFormation from %s", self.name, self.path
        )

        if not os.path.isfile(self.path):
            raise IOError("No such file or directory: '%s'", self.path)

        module = imp.load_source(self.name, self.path)

        try:
            body = module.sceptre_handler(self.sceptre_user_data)
        except AttributeError as e:
            if 'sceptre_handler' in str(e):
                raise TemplateSceptreHandlerError(
                    "The template does not have the required "
                    "'sceptre_handler(sceptre_user_data)' function."
                )
            else:
                raise e
        for directory in relpaths_to_add:
            sys.path.remove(os.path.join(os.getcwd(), directory))
        return body

    def upload_to_s3(
            self, region, bucket_name, key_prefix, environment_path,
            stack_name, connection_manager
    ):
        """
        Uploads the template to ``bucket_name`` and returns its URL.

        The template is uploaded with the key
        ``<key_prefix>/<region>/<environment_path>/<stack_name>-<timestamp>.template``.

        :param region: The AWS region to create the bucket in.
        :type region: str
        :param bucket_name: The name of the bucket to create.
        :type bucket_name: str
        :param key_prefix: A string to prefix to the key used to store the
            template in S3.
        :type key_prefix: str
        :param environment_path: The environment that the stack belongs to.
        :type env_path: str
        :param stack_name: The name of the stack that this template creates.
        :type stack_name: str
        :param connection_manager: The connection manager used to make
            AWS calls.
        :type connection_manager: sceptre.connection_manager.ConnectionManager
        :returns: The URL of the template object in S3.
        :rtype: str
        :raises: botocore.exceptions.ClientError

        """
        self.logger.debug("%s - Uploading template to S3...", self.name)

        with self._boto_s3_lock:
            if not self._bucket_exists(bucket_name, connection_manager):
                self._create_bucket(region, bucket_name, connection_manager)

        # Remove any leading or trailing slashes the user may have added.
        key_prefix = key_prefix.strip("/")

        template_key = "/".join([
            key_prefix,
            region,
            environment_path,
            "{stack_name}-{time_stamp}.template".format(
                stack_name=stack_name,
                time_stamp=datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S-%fZ")
            )
        ])

        self.logger.debug(
            "%s - Uploading template to: 's3://%s/%s'",
            self.name, bucket_name, template_key
        )
        connection_manager.call(
            service="s3",
            command="put_object",
            kwargs={
                "Bucket": bucket_name,
                "Key": template_key,
                "Body": self.body,
                "ServerSideEncryption": "AES256"
            }
        )

        url = "https://{0}.s3.amazonaws.com/{1}".format(
            bucket_name, template_key
        )

        self.logger.debug("%s - Template URL: '%s'", self.name, url)

        return url

    def _bucket_exists(self, bucket_name, connection_manager):
        """
        Checks if the bucket ``bucket_name`` exists.

        :param bucket_name: The name of the bucket to check.
        :type bucket_name: str
        :param connection_manager: The connection manager used to make
            AWS calls.
        :type connection_manager: sceptre.connection_manager.ConnectionManager
        :returns: Boolean whether the bucket exists
        :rtype: bool
        :raises: botocore.exception.ClientError

        """
        self.logger.debug(
            "%s - Attempting to find template bucket '%s'",
            self.name, bucket_name
        )
        try:
            connection_manager.call(
                service="s3",
                command="head_bucket",
                kwargs={"Bucket": bucket_name}
            )
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Message"] == "Not Found":
                self.logger.debug(
                    "%s - %s bucket not found.", self.name, bucket_name
                )
                return False
            else:
                raise
        self.logger.debug(
            "%s - Found template bucket '%s'", self.name, bucket_name
        )
        return True

    def _create_bucket(self, region, bucket_name, connection_manager):
        """
        Create the bucket ``bucket_name`` in the region ``region``.

        :param region: The AWS region to create the bucket in.
        :type region: str
        :param bucket_name: The name of the bucket to create.
        :type bucket_name: str
        :param connection_manager: The connection manager used to make
            AWS calls.
        :type connection_manager: sceptre.connection_manager.ConnectionManager
        :raises: botocore.exception.ClientError

        """
        self.logger.debug(
            "%s - Creating new bucket '%s'", self.name, bucket_name
        )
        if region == "us-east-1":
            connection_manager.call(
                service="s3",
                command="create_bucket",
                kwargs={"Bucket": bucket_name}
            )
        else:
            connection_manager.call(
                service="s3",
                command="create_bucket",
                kwargs={
                    "Bucket": bucket_name,
                    "CreateBucketConfiguration": {
                        "LocationConstraint": region
                    }
                }
            )

    @staticmethod
    def _render_jinja_template(template_dir, filename, jinja_vars):
        """
        Renders a jinja template.

        Sceptre supports passing sceptre_user_data to JSON and YAML
        CloudFormation templates using Jinja2 templating.

        :param template_dir: The directory containing the template.
        :type template_dir: str
        :param filename: The name of the template file.
        :type filename: str
        :param jinja_vars: Dict of variables to render into the template.
        :type jinja_vars: dict
        :returns: The body of the CloudFormation template.
        :rtype: string
        """
        logger = logging.getLogger(__name__)
        logger.debug("%s Rendering CloudFormation template", filename)
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir),
            undefined=jinja2.StrictUndefined
        )
        template = env.get_template(filename)
        body = template.render(**jinja_vars)
        return body



# -*- coding: utf-8 -*-

"""
sceptre.environment

This module implements the Environment class, which stores data and logic to
represent a logical grouping of stacks as an environment.

"""

from glob import glob
import logging
import os
import threading

import botocore

from concurrent.futures import ThreadPoolExecutor, wait

from .exceptions import StackDoesNotExistError

from .config import Config
from .connection_manager import ConnectionManager
from .exceptions import InvalidEnvironmentPathError
from .helpers import recurse_into_sub_environments, get_name_tuple
from .helpers import _detect_cycles
from .stack import Stack
from .stack_status import StackStatus


class Environment(object):
    """
    Environment stores information about the current environment.

    It implements methods for carrying out environment-level operations.

    Two types of Environments exist, non-leaf and leaf. Non-leaf environments
    contain sub-environments, while leaf environments contain stacks. If a
    command is executed by a leaf environment, it should execute that command
    on the stacks it contains. If a command is executed by a non-leaf
    environment, it should invoke that command on each of its sub-environments.
    This is done using the
    ``sceptre.helpers.recurse_into_sub_environments`` decorator.

    :param sceptre_dir: The absolute path to the Sceptre directory.
    :type project dir: str
    :param environment_path: The name of the environment.
    :type environment_path: str
    :param options: A dict of key-value pairs to update self.config with.
    :type debug: dict
    """
    def __init__(self, sceptre_dir, environment_path, options=None):
        self.logger = logging.getLogger(__name__)

        self.sceptre_dir = sceptre_dir
        self.path = self._validate_path(environment_path)
        self._options = {} if options is None else options

        self._is_leaf = None

        if self.is_leaf:
            self.stacks = self._load_stacks()
        else:
            self.environments = self._load_environments()

    def __repr__(self):
        return (
            "sceptre.environment.Environment(sceptre_dir='{0}', "
            "environment_path='{1}', options={2})".format(
                self.sceptre_dir, self.path, self._options
            )
        )

    @staticmethod
    def _validate_path(path):
        """
        Normalises backslashes to forward slashes for non-unix systems.
        Raises an InvalidEnvironmentPathError if the path has a leading or
        trailing slash.

        :param path: A directory path
        :type path: str
        :raises: sceptre.exceptions.InvalidEnvironmentPathError
        :returns: A normalised path with forward slashes.
        :rtype: string
        """
        path = path.replace("\\", "/")
        if path.endswith("/") or path.startswith("/"):
            raise InvalidEnvironmentPathError(
                "'{0}' is an invalid path string. Environment paths should "
                "not have leading or trailing slashes.".format(path)
            )
        return path

    @property
    def is_leaf(self):
        """
        Returns a boolean indicating if this environment is a leaf environment.

        The environment is a leaf environment if all items in the environent's
        directory are files (i.e. there are no sub-environments).

        :returns: A boolean indicating whether the environment is a leaf
            environment or not.
        :returns: bool
        """
        if self._is_leaf is None:
            self._is_leaf = all([
                os.path.isfile(path)
                for path in glob(os.path.join(
                    self.sceptre_dir, "config", self.path, "*"
                ))
            ])
        return self._is_leaf

    def launch(self):
        """
        Creates or updates all stacks in the environment.

        :returns: dict
        """
        self.logger.debug("Launching environment '%s'", self.path)
        threading_events = self._get_threading_events()
        stack_statuses = self._get_initial_statuses()
        launch_dependencies = self._get_launch_dependencies(self.path)

        self._check_for_circular_dependencies()
        self._build(
            "launch", threading_events, stack_statuses, launch_dependencies
        )
        return stack_statuses

    def delete(self):
        """
        Deletes all stacks in the environment.

        :returns: dict
        """
        self.logger.debug("Deleting environment '%s'", self.path)
        threading_events = self._get_threading_events()
        stack_statuses = self._get_initial_statuses()
        delete_dependencies = self._get_delete_dependencies()

        self._check_for_circular_dependencies()
        self._build(
            "delete", threading_events, stack_statuses, delete_dependencies
        )
        return stack_statuses

    @recurse_into_sub_environments
    def describe(self):
        """
        Returns each stack's status.

        :returns: The stack status of each stack, keyed by the stack's name.
        :rtype: dict
        """
        response = {}
        for stack in self.stacks.values():
            try:
                status = stack.get_status()
            except StackDoesNotExistError:
                status = "PENDING"
            response.update({stack.name: status})
        return response

    @recurse_into_sub_environments
    def describe_resources(self):
        """
        Describes the resources of each stack in the environment.

        :returns: A description of each stack's resources, keyed by the stack's
            name.
        :rtype: dict
        """
        response = {}
        for stack in self.stacks.values():
            try:
                resources = stack.describe_resources()
                response.update({stack.name: resources})
            except(botocore.exceptions.ClientError) as exp:
                if exp.response["Error"]["Message"].endswith("does not exist"):
                    pass
                else:
                    raise
        return response

    @recurse_into_sub_environments
    def _build(self, command, threading_events, stack_statuses, dependencies):
        """
        Launches or deletes all stacks in the environment.

        Whether the stack is launched or delete depends on the value of
        <command>. It does this by calling stack.<command>() for
        each stack in the environment. Stack.<command>() is blocking, because
        it waits for the stack to be built, so each command is run on a
        separate thread. As some stacks need to be built before others,
        depending on their depedencies, threading.Events() are used to notify
        the other stacks when a particular stack is done building.

        :param command: The stack command to run. Can be (launch | delete).
        :type command: str
        """
        if self.stacks:
            num_stacks = len(self.stacks)
            with ThreadPoolExecutor(max_workers=num_stacks) as executor:
                futures = [
                    executor.submit(
                        self._manage_stack_build, stack,
                        command, threading_events, stack_statuses, dependencies
                    )
                    for stack in self.stacks.values()
                ]
                wait(futures)
        else:
            self.logger.info(
                "No stacks found for environment: '%s'", self.path
            )

    def _manage_stack_build(
            self, stack, command, threading_events,
            stack_statuses, dependencies
    ):
        """
        Manages the launch or deletion of a stack.

        Waits for any stacks that ``stack`` depends on to complete, runs
        ``stack.command()``, and then marks ``stack`` as complete. If a
        dependency stack has failed, ``stack`` is marked as failed.

        :param stack: The stack to build.
        :type: sceptre.Stack
        :param threading_events: A dict of threading.Events, keyed by stack \
        names, which is used to notify other stacks when a particular stack \
        has been built.
        :type events: dict
        :param command: The stack command to run. Can be either "launch" or \
        "delete".
        :type command: str
        """
        for dependency in dependencies[stack.name]:
            threading_events[dependency].wait()
            if stack_statuses[dependency] != StackStatus.COMPLETE:
                self.logger.debug(
                    "%s, which %s depends is not complete. Marking "
                    "%s as failed.", dependency, stack.name, dependency
                )
                stack_statuses[stack.name] = StackStatus.FAILED

        if stack_statuses[stack.name] != StackStatus.FAILED:
            try:
                status = getattr(stack, command)()
                stack_statuses[stack.name] = status
            except Exception:
                stack_statuses[stack.name] = StackStatus.FAILED
                self.logger.exception(
                    "Stack %s failed to %s", stack.name, command
                )

        threading_events[stack.name].set()

    @recurse_into_sub_environments
    def _get_threading_events(self):
        """
        Returns a threading.Event() for each stack in every sub-environment.

        :returns: A threading.Event object for each stack, keyed by the
            stack's name.
        :rtype: dict
        """
        return {
            stack.name: threading.Event()
            for stack in self.stacks.values()
        }

    @recurse_into_sub_environments
    def _get_initial_statuses(self):
        """
        Returns a "pending" sceptre.stack_status.StackStatus for each stack
        in every sub-environment.

        :returns: A "pending" stack status for each stack, keyed by the
            stack's name.
        :rtype: dict
        """
        return {
            stack.name: StackStatus.PENDING
            for stack in self.stacks.values()
        }

    @recurse_into_sub_environments
    def _get_launch_dependencies(self, top_level_environment_path):
        """
        Returns a dict of each stack's launch dependencies.

        :returns: A list of the stacks that a particular stack depends on
            while launching, keyed by that stack's name.
        :rtype: dict
        """
        all_dependencies = {
            stack.name: stack.dependencies
            for stack in self.stacks.values()
        }

        # Filter out dependencies which aren't under the top level environmnent
        launch_dependencies = {
            stack_name: [
                dependency
                for dependency in dependencies
                if dependency.startswith(top_level_environment_path)
            ]
            for stack_name, dependencies in all_dependencies.items()
        }
        return launch_dependencies

    def _get_delete_dependencies(self):
        """
        Returns a dict of each stack's delete dependencies.

        :returns: A list of the stacks that a particular stack depends on
            while deleting, keyed by that stack's name.
        :rtype: dict
        """
        launch_dependencies = self._get_launch_dependencies(self.path)
        delete_dependencies = {
            stack_name: [] for stack_name in launch_dependencies
        }
        for stack_name, dependencies in launch_dependencies.items():
            for dependency in dependencies:
                delete_dependencies[dependency].append(stack_name)
        return delete_dependencies

    def _check_for_circular_dependencies(self):
        """
        Checks to make sure that no stacks are dependent on stacks which are
        dependent on the first stack.

        :raises: sceptre.workplan.CircularDependenciesException
        """
        self.logger.debug("Checking for circular dependencies...")

        if self.is_leaf:
            encountered_stacks = {}
            for stack in self.stacks.values():
                if encountered_stacks.get(stack, "UNENCOUNTERED") != "DONE":
                    encountered_stacks[stack] = "ENCOUNTERED"
                    encountered_stacks = _detect_cycles(
                        stack,
                        encountered_stacks,
                        self.stacks,
                        [stack.name]
                    )
                    encountered_stacks[stack] = "DONE"
        self.logger.debug("No circular dependencies found")

    def _get_config(self):
        """
        Initialises sceptre.Config and reads in the configuration data
        from each ``config.yaml`` file along the environment path. Overwrites
        the config with the values in ``self._options``.

        :returns: The environment's configuration.
        :rtype: dict
        """
        config = Config(
            sceptre_dir=self.sceptre_dir,
            environment_path=self.path,
            base_file_name="config"
        )
        config.read(self._options.get("user_variables"))
        config.update(self._options)

        return config

    def _get_available_stacks(self):
        """
        Returns all the stacks contained in the environment.

        :returns: A list of the environment's stacks.
        :rtype: list
        """
        config_files = glob(os.path.join(
            self.sceptre_dir, "config", self.path, "*.yaml"
        ))
        basenames = [
            os.path.splitext(os.path.basename(config_file))[0]
            for config_file in config_files
        ]

        if "config" in basenames:
            basenames.remove("config")

        stack_names = [
            "/".join([self.path, basename]) for basename in basenames
        ]
        return stack_names

    def _load_stacks(self):
        """
        Initialises and returns all of the environment's stacks.

        :returns: A dictionary of the stacks, keyed by the stack's name
        :rtype: dict
        """
        config = self._get_config()
        connection_manager = ConnectionManager(
            region=config["region"],
            iam_role=config.get("iam_role"),
            profile=config.get("profile")
        )
        stacks = {}
        for stack_name in self._get_available_stacks():
            self.logger.debug("Initialising stack '%s'", stack_name)
            stack = Stack(
                name=stack_name,
                environment_config=config,
                connection_manager=connection_manager
            )
            stacks[get_name_tuple(stack_name)[-1]] = stack
        return stacks

    def _get_available_environments(self):
        """
        Returns all the sub-environments contained in the environment.

        :returns: A list of the environment's sub-environments.
        :rtype: list
        """
        items_in_dir = glob(os.path.join(
            self.sceptre_dir, "config", self.path, "*"
        ))
        dirs = [
            item for item in items_in_dir
            if os.path.isdir(item)
        ]
        available_environments = [
            os.path.relpath(d, os.path.join(self.sceptre_dir, "config"))
            for d in dirs
        ]
        return available_environments

    def _load_environments(self):
        """
        Initialises and returns all of the environment's sub-environments.

        :returns: A dict of sub-environments, keyed by the environment's names
        :returns: dict
        """
        environments = {}

        for environment_name in self._get_available_environments():
            self.logger.debug(
                "Initialising '%s' environment...", environment_name
            )
            environment = Environment(
                sceptre_dir=self.sceptre_dir,
                environment_path=environment_name,
                options=self._options
            )
            environments[environment_name] = environment
        return environments



# -*- coding: utf-8 -*-

from contextlib import contextmanager
from tempfile import mkdtemp
import shutil
import os
from mock import patch, sentinel, call, Mock, ANY
import pytest

from sceptre.config import Config
from sceptre.hooks import Hook
from sceptre.resolvers import Resolver
from sceptre.exceptions import ConfigItemNotFoundError
from sceptre.exceptions import EnvironmentPathNotFoundError
from sceptre.exceptions import VersionIncompatibleError


class TestConfig(object):

    @patch("sceptre.config.Config._check_env_path_exists")
    def setup_method(self, test_method, mock_check_env_path_exists):
        self.config = Config(
            sceptre_dir="sceptre_dir",
            environment_path="environment_path",
            base_file_name="config"
        )

    @contextmanager
    def _create_temp_dir(self):
        temp_directory = mkdtemp()
        yield temp_directory
        shutil.rmtree(temp_directory)

    def test_config_correctly_initialised(self):
        assert self.config.sceptre_dir == "sceptre_dir"
        assert self.config.environment_path == "environment_path"
        assert self.config.name == "config"

    @patch("sceptre.config.Config.add_hook_constructors")
    @patch("sceptre.config.Config.add_resolver_constructors")
    @patch("sceptre.config.Config._check_env_path_exists")
    def test_initialise_with_yaml_constructors(
        self, mock_check_env_path_exists,
        mock_add_resolver_constructors, mock_add_hook_constructors
    ):
        config = Config.with_yaml_constructors(
            sceptre_dir="sceptre_dir",
            environment_path="environment_path",
            base_file_name="config",
            environment_config=sentinel.environment_config,
            connection_manager=sentinel.connection_manager
        )
        assert config.sceptre_dir == "sceptre_dir"
        assert config.environment_path == "environment_path"
        assert config.name == "config"
        mock_add_resolver_constructors.assert_called_once_with(
            sentinel.environment_config, sentinel.connection_manager
        )
        mock_add_hook_constructors.assert_called_once_with(
            sentinel.environment_config, sentinel.connection_manager
        )

    def test_get_attribute_with_valid_attribute(self):
        self.config["key"] = "value"
        assert self.config["key"] == "value"

    def test_get_attribute_with_invalid_attribute(self):
        with pytest.raises(ConfigItemNotFoundError):
            self.config["this_key_does_not_exist"]

    def test_read_reads_single_config_file(self):
        self.config.sceptre_dir = os.path.join(
            os.getcwd(), "tests", "fixtures"
        )
        self.config.environment_path = os.path.join(
            "account", "environment", "region"
        )
        self.config.name = "vpc"
        self.config.read()
        assert self.config == {
            'parameters': [{'param1': 'val1'}],
            'dependencies': [],
            'template_path': 'path/to/template'
        }

    def test_read_with_empty_config_file(self):
        self.config.sceptre_dir = os.path.join(
            os.getcwd(), "tests", "fixtures"
        )
        self.config.environment_path = os.path.join(
            "account", "environment", "region"
        )
        self.config.name = "subnets"

        self.config.read()
        assert self.config == {
            'dependencies': []
        }

    def test_read_cascades_config(self):
        self.config.sceptre_dir = os.path.join(
            os.getcwd(), "tests", "fixtures"
        )
        self.config.environment_path = os.path.join(
            "account", "environment", "region"
        )
        self.config.name = "config"
        self.config.read()
        assert self.config == {
            "region": "region_region",
            "template_bucket_name": "environment_template_bucket_name",
            "project_code": "account_project_code",
            "iam_role": "account_iam_role",
            "require_version": ">=0a",
            'dependencies': []
        }

    def test_read_with_templated_config_file(self):
        self.config.sceptre_dir = os.path.join(
            os.getcwd(), "tests", "fixtures"
        )
        self.config.environment_path = os.path.join(
            "account", "environment", "region"
        )
        self.config.name = "security_groups"
        os.environ["TEST_ENV_VAR"] = "environment_variable_value"
        user_variables = {"user_variable": "user_variable_value"}
        environment_config = {"region": "region"}
        self.config.read(user_variables, environment_config)

        assert self.config == {
            "parameters": {
                "param1": "user_variable_value",
                "param2": "environment_variable_value",
                "param3": "account",
                "param4": "environment",
                "param5": "region",
                "param6": "region"
            },
            'dependencies': []
        }

    def test_check_env_path_exists_with_valid_dir(self):
        with self._create_temp_dir() as temp_dir:
            self.config._check_env_path_exists(temp_dir)

    def test_check_env_path_exists_with_invalid_dir(self):
        with pytest.raises(EnvironmentPathNotFoundError):
            self.config._check_env_path_exists(
                "/this/directory/does/not/exist"
            )

    def test_aborts_on_incompatible_version_requirement(self):
        self.config['require_version'] = '<0'
        with pytest.raises(VersionIncompatibleError):
            self.config._check_version()

    @patch("sceptre.config.get_subclasses")
    @patch("sceptre.config.yaml.SafeLoader.add_constructor")
    def test_add_yaml_constructors(
        self, mock_add_constructors, mock_get_subclasses
    ):
        mock_get_subclasses.return_value = {
            "class_1": sentinel.class_1,
            "class_2": sentinel.class_2,
            "class_3": sentinel.class_3,
        }
        directory = "directory/path"
        base_type = str
        function = Mock(return_value="class")
        self.config.add_yaml_constructors(directory, base_type, function)
        calls = [call(u'!class_1', "class"),
                 call(u'!class_2', "class"),
                 call(u'!class_3', "class")]
        mock_add_constructors.assert_has_calls(calls, any_order=True)

    @patch("sceptre.config.os.path.dirname")
    @patch("sceptre.config.Config.add_yaml_constructors")
    def test_add_resolver_constructors(
        self, mock_add_yaml_constructors, mock_dirname
    ):
        mock_dirname.return_value = "folder/with/file"
        environment_config = sentinel.environment_config
        connection_manager = sentinel.connection_manager
        self.config.add_resolver_constructors(
            environment_config, connection_manager
        )
        calls = [
            call("folder/with/file/resolvers", Resolver, ANY),
            call("sceptre_dir/resolvers", Resolver, ANY)
        ]
        mock_add_yaml_constructors.assert_has_calls(calls, any_order=False)

    @patch("sceptre.config.os.path.dirname")
    @patch("sceptre.config.Config.add_yaml_constructors")
    def test_add_hook_constructors(
        self, mock_add_yaml_constructors, mock_dirname
    ):
        mock_dirname.return_value = "folder/with/file"
        environment_config = sentinel.environment_config
        connection_manager = sentinel.connection_manager
        self.config.add_hook_constructors(
            environment_config, connection_manager
        )
        calls = [
            call("folder/with/file/hooks", Hook, ANY),
            call("sceptre_dir/hooks", Hook, ANY)
        ]
        mock_add_yaml_constructors.assert_has_calls(calls, any_order=False)



# -*- coding: utf-8 -*-

"""
sceptre.cli

This module implements Sceptre's CLI, and should not be directly imported.
"""

import contextlib
import errno
from json import JSONEncoder
import os
import logging
from logging import Formatter
import sys
from uuid import uuid1
from functools import wraps
import warnings

import click
import colorama
import yaml
from boto3.exceptions import Boto3Error
from botocore.exceptions import BotoCoreError, ClientError
from jinja2.exceptions import TemplateError

from .config import ENVIRONMENT_CONFIG_ATTRIBUTES
from .environment import Environment
from .exceptions import SceptreException, ProjectAlreadyExistsError
from .stack_status import StackStatus, StackChangeSetStatus
from .stack_status_colourer import StackStatusColourer
from . import __version__


def environment_options(func):
    """
    environment_options is a decorator which adds the environment argument to
    the click function ``func``.

    :param func: The click function to add the arguments to.
    :type func: function
    :returns: function
    """
    func = click.argument("environment")(func)
    return func


def stack_options(func):
    """
    stack_options is a decorator which adds the stack and environment arguments
    to the click function ``func``.

    :param func: The click function to add the arguments to.
    :type func: function
    :returns: function
    """
    func = click.argument("stack")(func)
    func = click.argument("environment")(func)
    return func


def change_set_options(func):
    """
    change_set_options is a decorator which adds the environment, stack and
    change set name arguments to the click function ``func``.

    :param func: The click function to add the arguments to.
    :type func: function
    :returns: function
    """
    func = click.argument("change_set_name")(func)
    func = click.argument("stack")(func)
    func = click.argument("environment")(func)
    return func


def catch_exceptions(func):
    """
    Catches and simplifies expected errors thrown by sceptre.

    catch_exceptions should be used as a decorator.

    :param func: The function which may throw exceptions which should be
        simplified.
    :type func: func
    :returns: The decorated function.
    :rtype: func
    """
    @wraps(func)
    def decorated(*args, **kwargs):
        """
        Invokes ``func``, catches expected errors, prints the error message and
        exits sceptre with a non-zero exit code.
        """
        try:
            return func(*args, **kwargs)
        except (SceptreException, BotoCoreError, ClientError, Boto3Error,
                TemplateError) as error:
            write(error)
            sys.exit(1)

    return decorated


@click.group()
@click.version_option(version=__version__, prog_name="Sceptre")
@click.option("--debug", is_flag=True, help="Turn on debug logging.")
@click.option("--dir", "directory", help="Specify sceptre directory.")
@click.option(
    "--output", type=click.Choice(["yaml", "json"]), default="yaml",
    help="The formatting style for command output.")
@click.option("--no-colour", is_flag=True, help="Turn off output colouring.")
@click.option(
    "--var", multiple=True, help="A variable to template into config files.")
@click.option(
    "--var-file", type=click.File("rb"),
    help="A YAML file of variables to template into config files.")
@click.pass_context
def cli(
        ctx, debug, directory, no_colour, output, var, var_file
):  # pragma: no cover
    """
    Implements sceptre's CLI.
    """
    setup_logging(debug, no_colour)
    colorama.init()
    # Enable deprecation warnings
    warnings.simplefilter("always", DeprecationWarning)
    ctx.obj = {
        "options": {},
        "output_format": output,
        "no_colour": no_colour,
        "sceptre_dir": directory if directory else os.getcwd()
    }
    user_variables = {}
    if var_file:
        user_variables.update(yaml.safe_load(var_file.read()))
    if var:
        # --var options overwrite --var-file options
        for variable in var:
            variable_key, variable_value = variable.split("=")
            user_variables.update({variable_key: variable_value})
    if user_variables:
        ctx.obj["options"]["user_variables"] = user_variables


@cli.command(name="validate-template")
@stack_options
@click.pass_context
@catch_exceptions
def validate_template(ctx, environment, stack):
    """
    Validates the template.

    Validates ENVIRONMENT/STACK's template.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].validate_template()
    if response['ResponseMetadata']['HTTPStatusCode'] == 200:
        success_message = "Template is valid. Template details:\n"
        response = _remove_response_metadata(response)
        write(success_message, 'str')
        write(response, ctx.obj["output_format"])


@cli.command(name="generate-template")
@stack_options
@click.pass_context
@catch_exceptions
def generate_template(ctx, environment, stack):
    """
    Displays the template used.

    Prints ENVIRONMENT/STACK's template.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    template_output = env.stacks[stack].template.body
    write(template_output)


@cli.command(name="lock-stack")
@stack_options
@click.pass_context
@catch_exceptions
def lock_stack(ctx, environment, stack):
    """
    Prevents stack updates.

    Applies a stack policy to ENVIRONMENT/STACK which prevents updates.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].lock()


@cli.command(name="unlock-stack")
@stack_options
@click.pass_context
@catch_exceptions
def unlock_stack(ctx, environment, stack):
    """
    Allows stack updates.

    Applies a stack policy to ENVIRONMENT/STACK which allows updates.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].unlock()


# DESCRIBE RESOURCES should merge into "describe-resources"
@cli.command(name="describe-env-resources")
@environment_options
@click.pass_context
@catch_exceptions
def describe_env_resources(ctx, environment):
    """
    Describes the env's resources.

    Prints ENVIRONMENT's resources.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    responses = env.describe_resources()

    write(responses, ctx.obj["output_format"])


@cli.command(name="describe-stack-resources")
@stack_options
@click.pass_context
@catch_exceptions
def describe_stack_resources(ctx, environment, stack):
    """
    Describes the stack's resources.

    Prints ENVIRONMENT/STACK's resources.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].describe_resources()

    write(response, ctx.obj["output_format"])


@cli.command(name="create-stack")
@stack_options
@click.pass_context
@catch_exceptions
def create_stack(ctx, environment, stack):
    """
    Creates the stack.

    Creates ENVIRONMENT/STACK.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].create()
    if response != StackStatus.COMPLETE:
        exit(1)


@cli.command(name="delete-stack")
@stack_options
@click.pass_context
@catch_exceptions
def delete_stack(ctx, environment, stack):
    """Deletes the stack.

    Deletes ENVIRONMENT/STACK."""
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].delete()
    if response != StackStatus.COMPLETE:
        exit(1)


@cli.command(name="update-stack")
@stack_options
@click.pass_context
@catch_exceptions
def update_stack(ctx, environment, stack):
    """
    Updates the stack.

    Updates ENVIRONMENT/STACK.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].update()
    if response != StackStatus.COMPLETE:
        exit(1)


@cli.command(name="launch-stack")
@stack_options
@click.pass_context
@catch_exceptions
def launch_stack(ctx, environment, stack):
    """
    Creates or updates the stack.

    Creates or updates ENVIRONMENT/STACK.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].launch()
    if response != StackStatus.COMPLETE:
        exit(1)


@cli.command(name="launch-env")
@environment_options
@click.pass_context
@catch_exceptions
def launch_env(ctx, environment):
    """
    Creates or updates all stacks.

    Creates or updates all the stacks in ENVIRONMENT.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.launch()
    if not all(status == StackStatus.COMPLETE for status in response.values()):
        exit(1)


@cli.command(name="delete-env")
@environment_options
@click.pass_context
@catch_exceptions
def delete_env(ctx, environment):
    """
    Deletes all stacks.

    Deletes all the stacks in ENVIRONMENT.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.delete()
    if not all(status == StackStatus.COMPLETE for status in response.values()):
        exit(1)


@cli.command(name="continue-update-rollback")
@stack_options
@click.pass_context
@catch_exceptions
def continue_update_rollback(ctx, environment, stack):
    """
    Rolls stack back to working state.

    If ENVIRONMENT/STACK is in the state UPDATE_ROLLBACK_FAILED, roll it
    back to the UPDATE_ROLLBACK_COMPLETE state.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].continue_update_rollback()


@cli.command(name="create-change-set")
@change_set_options
@click.pass_context
@catch_exceptions
def create_change_set(ctx, environment, stack, change_set_name):
    """
    Creates a change set.

    Create a change set for ENVIRONMENT/STACK with the name
    CHANGE_SET_NAME.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].create_change_set(change_set_name)


@cli.command(name="delete-change-set")
@change_set_options
@click.pass_context
@catch_exceptions
def delete_change_set(ctx, environment, stack, change_set_name):
    """
    Deletes the change set.

    Deletes ENVIRONMENT/STACK's change set with name CHANGE_SET_NAME.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].delete_change_set(change_set_name)


@cli.command(name="describe-change-set")
@change_set_options
@click.option("--verbose", is_flag=True)
@click.pass_context
@catch_exceptions
def describe_change_set(ctx, environment, stack, change_set_name, verbose):
    """
    Describes the change set.

    Describes ENVIRONMENT/STACK's change set with name CHANGE_SET_NAME.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    description = env.stacks[stack].describe_change_set(change_set_name)
    if not verbose:
        description = _simplify_change_set_description(description)
    write(description, ctx.obj["output_format"])


def _simplify_change_set_description(response):
    desired_response_items = [
        "ChangeSetName",
        "CreationTime",
        "ExecutionStatus",
        "StackName",
        "Status",
        "StatusReason"
    ]
    desired_resource_changes = [
        "Action",
        "LogicalResourceId",
        "PhysicalResourceId",
        "Replacement",
        "ResourceType",
        "Scope"
    ]
    formatted_response = {
        k: v
        for k, v in response.items()
        if k in desired_response_items
    }
    formatted_response["Changes"] = [
        {
            "ResourceChange": {
                k: v
                for k, v in change["ResourceChange"].items()
                if k in desired_resource_changes
            }
        }
        for change in response["Changes"]
    ]
    return formatted_response


@cli.command(name="execute-change-set")
@change_set_options
@click.pass_context
@catch_exceptions
def execute_change_set(ctx, environment, stack, change_set_name):
    """
    Executes the change set.

    Executes ENVIRONMENT/STACK's change set with name CHANGE_SET_NAME.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].execute_change_set(change_set_name)


@cli.command(name="list-change-sets")
@stack_options
@click.pass_context
@catch_exceptions
def list_change_sets(ctx, environment, stack):
    """
    Lists change sets.

    Lists ENVIRONMENT/STACK's change sets.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].list_change_sets()
    formatted_response = {
        k: v
        for k, v in response.items()
        if k != "ResponseMetadata"
    }
    write(formatted_response, ctx.obj["output_format"])


@cli.command(name="update-stack-cs")
@stack_options
@click.option("--verbose", is_flag=True)
@click.pass_context
@catch_exceptions
def update_with_change_set(ctx, environment, stack, verbose):
    """
    Updates the stack using a change set.

    Creates a change set for ENVIRONMENT/STACK, prints out a description of the
    changes, and prompts the user to decide whether to execute or delete it.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    change_set_name = "-".join(["change-set", uuid1().hex])
    with change_set(env.stacks[stack], change_set_name):
        status = env.stacks[stack].wait_for_cs_completion(change_set_name)
        description = env.stacks[stack].describe_change_set(change_set_name)
        if not verbose:
            description = _simplify_change_set_description(description)
        write(description, ctx.obj["output_format"])
        if status != StackChangeSetStatus.READY:
            exit(1)
        if click.confirm("Proceed with stack update?"):
            env.stacks[stack].execute_change_set(change_set_name)


@contextlib.contextmanager
def change_set(stack, name):
    """
    Creates and yields and deletes a change set.

    :param stack: The stack to create the change set for.
    :type stack: sceptre.stack.Stack
    :param name: The name of the change set.
    :type name: str
    """
    stack.create_change_set(name)
    try:
        yield
    finally:
        stack.delete_change_set(name)


@cli.command(name="describe-stack-outputs")
@stack_options
@click.option("--export", type=click.Choice(["envvar"]))
@click.pass_context
@catch_exceptions
def describe_stack_outputs(ctx, environment, stack, export):
    """
    Describes stack outputs.

    Describes ENVIRONMENT/STACK's stack outputs.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].describe_outputs()

    if export == "envvar":
        write("\n".join(
            [
                "export SCEPTRE_{0}={1}".format(
                    output["OutputKey"], output["OutputValue"]
                )
                for output in response
            ]
        ))
    else:
        write(response, ctx.obj["output_format"])


@cli.command(name="describe-env")
@environment_options
@click.pass_context
@catch_exceptions
def describe_env(ctx, environment):
    """
    Describes the stack statuses.

    Describes ENVIRONMENT stack statuses.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    responses = env.describe()
    write(responses, ctx.obj["output_format"], ctx.obj["no_colour"])


@cli.command(name="set-stack-policy")
@stack_options
@click.option("--policy-file")
@click.pass_context
@catch_exceptions
def set_stack_policy(ctx, environment, stack, policy_file):
    """
    Sets stack policy.

    Sets a specific ENVIRONMENT/STACK policy.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    env.stacks[stack].set_policy(policy_file)


@cli.command(name="get-stack-policy")
@stack_options
@click.pass_context
@catch_exceptions
def get_stack_policy(ctx, environment, stack):
    """
    Displays the stack policy used.

    Prints ENVIRONMENT/STACK policy.
    """
    env = get_env(ctx.obj["sceptre_dir"], environment, ctx.obj["options"])
    response = env.stacks[stack].get_policy()

    write(response.get('StackPolicyBody', {}))


@cli.group(name="init")
def init():
    """
    Commands for initialising Sceptre projects.
    """
    pass


@init.command("env")
@click.argument('env')
@catch_exceptions
@click.pass_context
def init_environment(ctx, env):
    """
    Initialises an environment in a project.

    Creates ENVIRONMENT folder in the project.
    """
    cwd = ctx.obj["sceptre_dir"]
    for item in os.listdir(cwd):
        # If already a config folder create a sub environment
        if os.path.isdir(item) and item == "config":
            config_dir = os.path.join(os.getcwd(), "config")
            _create_new_environment(config_dir, env)


@init.command("project")
@catch_exceptions
@click.argument('project_name')
@click.pass_context
def init_project(ctx, project_name):
    """
    Initialises a new project.

    Creates PROJECT_NAME project folder.
    """
    cwd = os.getcwd()
    sceptre_folders = {"config", "templates"}
    project_folder = os.path.join(cwd, project_name)
    try:
        os.mkdir(project_folder)
    except OSError as e:
        # Check if environment folder already exists
        if e.errno == errno.EEXIST:
            raise ProjectAlreadyExistsError(
                'Folder \"{0}\" already exists.'.format(project_name)
            )
        else:
            raise

    for folder in sceptre_folders:
        folder_path = os.path.join(project_folder, folder)
        os.makedirs(folder_path)

    defaults = {
        "project_code": project_name,
        "region": os.environ.get("AWS_DEFAULT_REGION", "")
    }

    config_path = os.path.join(cwd, project_name, "config")
    _create_config_file(config_path, config_path, defaults)


def _create_new_environment(config_dir, new_path):
    """
    Creates the subfolder for the environment specified by `path` starting
    from the `config_dir`. Even if folder path already exists, ask the user if
    they want to initialise `config.yaml`.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param path: The directory path to the environment folder.
    :type path: str
    """
    # Create full path to environment
    folder_path = os.path.join(config_dir, new_path)
    init_config_msg = 'Do you want initialise config.yaml?'

    # Make folders for the environment
    try:
        os.makedirs(folder_path)
    except OSError as e:
        # Check if environment folder already exists
        if e.errno == errno.EEXIST:
            init_config_msg = 'Environment path exists. ' + init_config_msg
        else:
            raise

    if click.confirm(init_config_msg):
        _create_config_file(config_dir, folder_path)


def _get_nested_config(config_dir, path):
    """
    Collects nested config from between `config_dir` and `path`. Config at
    lower level as greater precedence.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param path: The directory path to the environment folder.
    :type path: str
    :returns: The nested config.
    :rtype: dict
    """
    config = {}
    for root, _, files in os.walk(config_dir):
        # Check that folder is within the final environment path
        if path.startswith(root) and "config.yaml" in files:
            config_path = os.path.join(root, "config.yaml")
            with open(config_path) as config_file:
                config.update(yaml.safe_load(config_file))
    return config


def _create_config_file(config_dir, path, defaults={}):
    """
    Creates a `config.yaml` file in the given path. The user is asked for
    values for requried properties. Defaults are suggested with values in
    `defaults` and then vaules found in parent `config.yaml` files. If
    properties and their values are the same as in parent `config.yaml`, then
    they are not included. No file is produced if require values are satisfied
    by parent `config.yaml` files.

    :param config_dir: The directory path to the top-level config folder.
    :type config_dir: str
    :param path: The directory path to the environment folder.
    :type path: str
    :param defaults: Defaults to present to the user for config.
    :type defaults: dict
    """
    config = dict.fromkeys(ENVIRONMENT_CONFIG_ATTRIBUTES.required, "")
    parent_config = _get_nested_config(config_dir, path)

    # Add standard defaults
    config.update(defaults)

    # Add parent config values as defaults
    config.update(parent_config)

    # Ask for new values
    for key, value in config.items():
        config[key] = click.prompt(
            'Please enter a {0}'.format(key), default=value
        )

    # Remove values if parent config are the same
    config = {k: v for k, v in config.items() if parent_config.get(k) != v}

    # Write config.yaml if config not empty
    filepath = os.path.join(path, "config.yaml")
    if config:
        with open(filepath, 'w') as config_file:
            yaml.safe_dump(
                config, stream=config_file, default_flow_style=False
            )
    else:
        click.echo("No config.yaml file needed - covered by parent config.")


def get_env(sceptre_dir, environment_path, options):
    """
    Initialises and returns a sceptre.environment.Environment().

    :param sceptre_dir: The absolute path to the Sceptre directory.
    :type project dir: str
    :param environment_path: The name of the environment.
    :type environment_path: str
    :param options: A dict of key-value pairs to update self.config with.
    :type debug: dict
    :returns: An Environment.
    :rtype: sceptre.environment.Environment
    """
    return Environment(
        sceptre_dir=sceptre_dir,
        environment_path=environment_path,
        options=options
    )


def setup_logging(debug, no_colour):
    """
    Sets up logging.

    By default, the python logging module is configured to push logs to stdout
    as long as their level is at least INFO. The log format is set to
    "[%(asctime)s] - %(name)s - %(message)s" and the date format is set to
    "%Y-%m-%d %H:%M:%S".

    After this function has run, modules should:

    .. code:: python

        import logging

        logging.getLogger(__name__).info("my log message")

    :param debug: A flag indication whether to turn on debug logging.
    :type debug: bool
    :no_colour: A flag to indicating whether to turn off coloured output.
    :type no_colour: bool
    :returns: A logger.
    :rtype: logging.Logger
    """
    if debug:
        sceptre_logging_level = logging.DEBUG
        logging.getLogger("botocore").setLevel(logging.INFO)
    else:
        sceptre_logging_level = logging.INFO
        # Silence botocore logs
        logging.getLogger("botocore").setLevel(logging.CRITICAL)

    formatter_class = Formatter if no_colour else ColouredFormatter

    formatter = formatter_class(
        fmt="[%(asctime)s] - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S"
    )

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    logger = logging.getLogger("sceptre")
    logger.addHandler(log_handler)
    logger.setLevel(sceptre_logging_level)
    return logger


def write(var, output_format="str", no_colour=True):
    """
    Writes ``var`` to stdout. If output_format is set to "json" or "yaml",
    write ``var`` as a JSON or YAML string.

    :param var: The object to print
    :type var: obj
    :param output_format: The format to print the output as. Allowed values: \
    "str", "json", "yaml"
    :type output_format: str
    :param no_colour: Whether to color stack statuses
    :type no_colour: bool
    """
    if output_format == "json":
        encoder = CustomJsonEncoder()
        stream = encoder.encode(var)
    if output_format == "yaml":
        stream = yaml.safe_dump(var, default_flow_style=False)
    if output_format == "str":
        stream = var

    if not no_colour:
        stack_status_colourer = StackStatusColourer()
        stream = stack_status_colourer.colour(stream)

    click.echo(stream)


def _remove_response_metadata(response):
    """
    Removes the Response Metadata from an AWS API reponse.

    :param response: The AWS response that you wish to remove the Response \
    Metadata from.
    :type response: dict
    :returns: The original dict with the Response Metadata removed.
    :rtype: dict
    """
    if "ResponseMetadata" in response:
        del response['ResponseMetadata']
    return response


class ColouredFormatter(Formatter):
    """
    ColouredFormatter add colours to all stack statuses that appear in log
    messages.
    """

    stack_status_colourer = StackStatusColourer()

    def format(self, record):
        """
        Colours and returns all stack statuses in ``record``.

        :param record: The log item to format.
        :type record: str
        :returns: str
        """
        response = super(ColouredFormatter, self).format(record)
        coloured_response = self.stack_status_colourer.colour(response)
        return coloured_response


class CustomJsonEncoder(JSONEncoder):
    """
    CustomJsonEncoder is a JSONEncoder which encodes all items as JSON by
    calling their __str__() method.
    """

    def default(self, item):
        """
        Returns stringified version of item.

        :param item: An arbitrary object to stringify.
        :type item: obj
        :returns: The stringified object.
        :rtype: str
        """
        return str(item)



# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock, Mock, patch, sentinel

from sceptre.config import Config
from sceptre.resolvers.stack_output import \
    StackOutput, StackOutputExternal, StackOutputBase
from sceptre.exceptions import DependencyStackMissingOutputError
from sceptre.exceptions import StackDoesNotExistError
from botocore.exceptions import ClientError


class TestStackOutputResolver(object):

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolve(self, mock_get_output_value):
        mock_environment_config = MagicMock(spec=Config)
        mock_stack_config = MagicMock(spec=Config)
        mock_environment_config.__getitem__.return_value = "project-code"
        mock_stack_config.__getitem__.return_value = []

        stack_output_resolver = StackOutput(
            environment_config=mock_environment_config,
            stack_config=mock_stack_config,
            connection_manager=sentinel.connection_manager,
            argument="account/dev/vpc::VpcId"
        )

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver.resolve()

        mock_get_output_value.assert_called_once_with(
            "project-code-account-dev-vpc", "VpcId"
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutput._get_output_value"
    )
    def test_resolve_with_implicit_stack_reference(
        self, mock_get_output_value
    ):
        mock_environment_config = MagicMock(spec=Config)
        mock_stack_config = MagicMock(spec=Config)
        mock_environment_config.__getitem__.return_value = "project-code"
        mock_stack_config.environment_path = "account/dev"
        mock_stack_config.__getitem__.return_value = []

        stack_output_resolver = StackOutput(
            environment_config=mock_environment_config,
            stack_config=mock_stack_config,
            connection_manager=sentinel.connection_manager,
            argument="vpc::VpcId"
        )

        mock_get_output_value.return_value = "output_value"

        stack_output_resolver.resolve()

        mock_get_output_value.assert_called_once_with(
            "project-code-account-dev-vpc", "VpcId"
        )


class TestStackOutputExternalResolver(object):

    def setup_method(self, test_method):
        self.stack_output_resolver = StackOutputExternal(
            environment_config=sentinel.environment_config,
            stack_config=sentinel.config,
            connection_manager=sentinel.connection_manager,
            argument=None
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutputExternal._get_output_value"
    )
    def test_resolve(self, mock_get_output_value):
        mock_environment_config = MagicMock(spec=Config)
        mock_environment_config.__getitem__.return_value = "project-code"
        mock_environment_config.environment_path = "account/dev"
        self.stack_output_resolver.environment_config = mock_environment_config
        self.stack_output_resolver.argument = "vpc::VpcId"

        mock_get_output_value.return_value = "output_value"

        self.stack_output_resolver.resolve()

        mock_get_output_value.assert_called_once_with(
            "vpc", "VpcId"
        )


class MockStackOutputBase(StackOutputBase):
    """
    MockBaseResolver inherits from the abstract base class
    StackOutputBaseResolver, and implements the abstract methods. It is used
    to allow testing on StackOutputBaseResolver, which is not otherwise
    instantiable.
    """
    def __init__(self, *args, **kwargs):
        super(MockStackOutputBase, self).__init__(*args, **kwargs)

    def resolve(self):
        pass


class TestStackOutputBaseResolver(object):

    def setup_method(self, test_method):
        self.mock_base_stack_output_resolver = MockStackOutputBase(
            environment_config=sentinel.environment_config,
            stack_config=sentinel.config,
            connection_manager=sentinel.connection_manager,
            argument=None
        )

    @patch(
        "sceptre.resolvers.stack_output.StackOutputBase._get_stack_outputs"
    )
    def test_get_output_value_with_valid_key(self, mock_get_stack_outputs):
        mock_get_stack_outputs.return_value = {"key": "value"}

        response = self.mock_base_stack_output_resolver._get_output_value(
            sentinel.stack_name, "key"
        )

        assert response == "value"

    @patch(
        "sceptre.resolvers.stack_output.StackOutputBase._get_stack_outputs"
    )
    def test_get_output_value_with_invalid_key(self, mock_get_stack_outputs):
        mock_get_stack_outputs.return_value = {"key": "value"}

        with pytest.raises(DependencyStackMissingOutputError):
            self.mock_base_stack_output_resolver._get_output_value(
                sentinel.stack_name, "invalid_key"
            )

    def test_get_stack_outputs_with_valid_stack(self):
        mock_connection_manager = Mock()
        mock_connection_manager.call.return_value = {
            "Stacks": [{
                "Outputs": [
                    {
                        "OutputKey": "key_1",
                        "OutputValue": "value_1",
                        "Description": "description_1"
                    },
                    {
                        "OutputKey": "key_2",
                        "OutputValue": "value_2",
                        "Description": "description_2"
                    }
                ]
            }]
        }
        self.mock_base_stack_output_resolver.connection_manager = \
            mock_connection_manager

        response = self.mock_base_stack_output_resolver._get_stack_outputs(
            sentinel.stack_name
        )
        assert response == {
            "key_1": "value_1",
            "key_2": "value_2"
        }

    def test_get_stack_outputs_with_unlaunched_stack(self):
        mock_connection_manager = Mock()
        mock_connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "404",
                    "Message": "stack does not exist"
                }
            },
            sentinel.operation
        )
        self.mock_base_stack_output_resolver.connection_manager = \
            mock_connection_manager

        with pytest.raises(StackDoesNotExistError):
            self.mock_base_stack_output_resolver._get_stack_outputs(
                sentinel.stack_name
            )

    def test_get_stack_outputs_with_unkown_boto_error(self):
        mock_connection_manager = Mock()
        mock_connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "500",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        self.mock_base_stack_output_resolver.connection_manager = \
            mock_connection_manager

        with pytest.raises(ClientError):
            self.mock_base_stack_output_resolver._get_stack_outputs(
                sentinel.stack_name
            )



# -*- coding: utf-8 -*-
import abc
import six
import logging


@six.add_metaclass(abc.ABCMeta)
class Resolver():
    """
    Resolver is an abstract base class that should be inherited by all
    resolvers. Environment config, stack config and the connection
    manager are supplied to the class, as they may be of use to inheriting
    classes.

    :param environment_config: The environment_config from config.yaml files.
    :type environment_config: sceptre.config.Config
    :param stack_config: The stack config.
    :type stack_config: sceptre.config.Config
    :param connection_manager: A connection manager.
    :type connection_manager: sceptre.connection_manager.ConnectionManager
    :param argument: Arguments to pass to the resolver.
    :type argument: str
    """

    def __init__(
        self, argument=None, connection_manager=None,
        environment_config=None, stack_config=None
    ):
        self.logger = logging.getLogger(__name__)
        self.environment_config = environment_config
        self.stack_config = stack_config
        self.connection_manager = connection_manager
        self.argument = argument

    @abc.abstractmethod
    def resolve(self):
        """
        An abstract method which must be overwritten by all inheriting classes.
        This method is called to retrieve the final desired value.
        Implementation of this method in subclasses must return a suitable
        object or primitive type.
        """
        pass  # pragma: no cover


class ResolvableProperty(object):
    """
    This is a descriptor class used to store an attribute that may contain
    Resolver objects. When retrieving the dictionary or list, any Resolver
    objects contains are a value or within a list are resolved to a primitive
    type. Supports nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: string
    """
    def __init__(self, name):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)

    def __get__(self, instance, type):
        """
        Attribute getter which resolves any Resolver object contained in the
        complex data structure.

        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: dict or list
        """
        if not hasattr(instance, self.name) \
                or getattr(instance, self.name) is None:
            value = instance.config.get(self.name[1:], {})
            setattr(instance, self.name, value)
        return self.resolve_values(getattr(instance, self.name))

    def __set__(self, instance, value):
        setattr(instance, self.name, value)

    def resolve_values(self, attr):
        """
        Searches through dictionary or list for Resolver objects and replaces
        them with the resolved value. Supports nested dictionaries and lists.
        Does not detect Resolver objects used as keys in dictionaries.

        :param attr: A complex data structure to search through.
        :type attr: dict or list
        :return: A complex data structure without Resolver objects.
        :rtype: dict or list
        """
        if isinstance(attr, dict):
            for key, value in attr.items():
                if isinstance(value, Resolver):
                    attr[key] = value.resolve()
                elif isinstance(value, list) or isinstance(value, dict):
                    self.resolve_values(value)
        elif isinstance(attr, list):
            for index, value in enumerate(attr):
                if isinstance(value, Resolver):
                    attr[index] = value.resolve()
                elif isinstance(value, list) or isinstance(value, dict):
                    self.resolve_values(value)
        return attr



# -*- coding: utf-8 -*-

import pytest
from colorama import Fore, Style
from mock import patch

from sceptre.hooks.bash import Bash
from sceptre.exceptions import InvalidHookArgumentTypeError


class TestBash(object):
    def setup_method(self, test_method):
        self.mock_bash = Bash()

    @patch('warnings.warn')
    def test_run_with_non_str_argument(self, mock_warn):
        self.mock_bash.argument = None
        with pytest.raises(InvalidHookArgumentTypeError):
            self.mock_bash.run()
        mock_warn.assert_called_once_with(
            "{0}The bash hook has been deprecated, "
            "and will be removed in a later version of "
            "Sceptre. Use the cmd hook instead. "
            "Example: !cmd <command>{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )

    @patch('warnings.warn')
    @patch('sceptre.hooks.bash.subprocess.call')
    def test_run_with_str_argument(self, mock_call, mock_warn):
        Bash.ALLOW_COMMAND_ERROR = True
        self.mock_bash.argument = u'echo hello'
        self.mock_bash.run()
        mock_call.assert_called_once_with(['/bin/bash', '-c', 'echo hello'])
        mock_warn.assert_called_once_with(
            "{0}The bash hook has been deprecated, "
            "and will be removed in a later version of "
            "Sceptre. Use the cmd hook instead. "
            "Example: !cmd <command>{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )

    @patch('warnings.warn')
    @patch("sceptre.hooks.bash.subprocess.check_call")
    def test_bash_deny_command_error(self, mock_check_call, mock_warn):
        Bash.ALLOW_COMMAND_ERROR = False
        self.mock_bash.argument = u'echo hello'
        self.mock_bash.run()
        mock_check_call.assert_called_once_with(
            ['/bin/bash', '-c', 'echo hello'])
        mock_warn.assert_called_once_with(
            "{0}The bash hook has been deprecated, "
            "and will be removed in a later version of "
            "Sceptre. Use the cmd hook instead. "
            "Example: !cmd <command>{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )



# -*- coding: utf-8 -*-

"""
sceptre.config

This module implements a Config class, which stores a stack or environment's
configuration.
"""

import logging
import os
import yaml
import collections

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from . import __version__
from .exceptions import ConfigItemNotFoundError
from .exceptions import EnvironmentPathNotFoundError
from .exceptions import VersionIncompatibleError
from .hooks import Hook
from .resolvers import Resolver
from .helpers import get_subclasses

ConfigAttributes = collections.namedtuple("Attributes", "required optional")

ENVIRONMENT_CONFIG_ATTRIBUTES = ConfigAttributes(
    {
        "project_code",
        "region"
    },
    {
        "iam_role",
        "template_bucket_name",
        "template_key_prefix",
        "require_version"
    }
)

STACK_CONFIG_ATTRIBUTES = ConfigAttributes(
    {
        "template_path"
    },
    {
        "dependencies",
        "hooks",
        "parameters",
        "protect",
        "sceptre_user_data",
        "stack_name",
        "stack_tags",
        "role_arn"
    }
)


class Config(dict):
    """
    Config stores the configuration read in from the YAML files. Config
    inherits from dict, and so configuration data can be accessed from Config
    as if Config were a dict. Config implements read(), a method to read in
    config from the ``<base_file_name>.yaml`` files along the
    ``environment_path`` from ``sceptre_dir``.

    :param sceptre_dir: The absolute path to the Sceptre directory.
    :type project dir: str
    :param environment_path: The name of the environment.
    :type environment_path: str
    :param base_file_name: The basename of the file to read in \
        (e.g. "config", "vpc")
    :type base_file_name: str
    """

    def __init__(self, sceptre_dir, environment_path, base_file_name):
        self.logger = logging.getLogger(__name__)

        self.sceptre_dir = sceptre_dir
        self.environment_path = environment_path
        self.name = base_file_name
        self["dependencies"] = []

        self._check_env_path_exists(os.path.join(
            self.sceptre_dir, "config", self.environment_path
        ))

        super(Config, self).__init__(self)

    @classmethod
    def with_yaml_constructors(
            cls, sceptre_dir, environment_path, base_file_name,
            environment_config, connection_manager
    ):
        """
        Initialises a Config object with additional PyYAML constructor classes.

        Creates a Config object and adds PyYAML constructor meta-classes.
        Additional constructors are added to transform YAML nodes for any
        library or external Resolvers and Hooks.

        :param sceptre_dir: The absolute path to the Sceptre directory.
        :type project dir: str
        :param environment_path: The name of the environment.
        :type environment_path: str
        :param base_file_name: The basename of the file to read in \
            (e.g. "config", "vpc")
        :type base_file_name: str
        :param environment_config: The stack's environment's config.
        :type environment_config: sceptre.config.Config
        :param connection_manager: A ConnectionManager used to make Boto3
            calls.
        :type connection_manager: sceptre.connection_manager.ConnectionManager
        """
        obj = cls(sceptre_dir, environment_path, base_file_name)
        obj.add_resolver_constructors(
            environment_config, connection_manager)
        obj.add_hook_constructors(
            environment_config, connection_manager)
        return obj

    def __getitem__(self, item):
        """
        Makes a call to dict's __getitem__. If the item isn't in the dict,
        catch the KeyError raised, and raise a ConfigItemNotFoundException,
        which adds more detail to the exception.

        :param item: The key of an item in the dict to return.
        :type item: Any hashable type
        :returns: The value of the config item keyed by ``item``.
        :rtype: obj
        :raises: sceptre.exceptions.ConfigItemNotFoundException
        """
        try:
            return super(Config, self).__getitem__(item)
        except KeyError:
            raise ConfigItemNotFoundError(
                "'{0}' was not found in any of the {1}.yaml files.".format(
                    item, self.name
                )
            )

    def read(self, user_variables=None, environment_config=None):
        """
        Reads in configuration from files.

        Traverses the environment path, from top to bottom, reading in all
        relevant config files. If config items appear in files lower down the
        environment tree, they overwrite items from further up. Jinja2 is used
        to template in variables from user_variables, environment variables,
        and the segments of the environment path.

        :param user_variables: A dict of key value pairs to be supplied to \
        the config file via Jinja2 templating.
        :type user_variables: dict
        """
        path = os.path.join("config", self.environment_path)
        file_name = ".".join([self.name, "yaml"])

        self.logger.debug("Reading in '%s' files...", file_name)

        def get_config(path):
            """
            Recursively reads in config files in nested subfolders.
            """
            if not path:
                return {}
            else:
                directory_path = os.path.join(self.sceptre_dir, path)
                config = {}
                if os.path.isfile(os.path.join(directory_path, file_name)):
                    env = Environment(
                        loader=FileSystemLoader(directory_path),
                        undefined=StrictUndefined
                    )
                    template = env.get_template(file_name)
                    rendered_template = template.render(
                        environment_variable=os.environ,
                        var=user_variables,
                        environment_path=self.environment_path.split("/"),
                        environment_config=environment_config
                    )

                    yaml_data = yaml.safe_load(rendered_template)
                    if yaml_data is not None:
                        config = yaml_data
                cascaded_config = get_config(os.path.dirname(path))
                cascaded_config.update(config)
                return cascaded_config

        config = get_config(path)

        self.update(config)

        self._check_version()

        self.logger.debug("Config: %s", self)

    @staticmethod
    def _check_env_path_exists(path):
        """
        Raises an EnvironmentPathNotFoundError if ``path`` is not a directory.

        :param path: A directory path.
        :type path: str
        :raises: sceptre.exceptions.EnvironmentPathNotFoundError
        """
        if not os.path.isdir(path):
            raise EnvironmentPathNotFoundError(
                "The environment '{0}' does not exist.".format(path)
            )

    def _check_version(self):
        """
        Raises a VersionIncompatibleException when the current sceptre version
        does not comply with the configured version requirement.

        :raises: sceptre.exceptions.VersionIncompatibleException
        """
        sceptre_version = __version__
        if self.name == 'config' and 'require_version' in self:
            require_version = self['require_version']
            if Version(sceptre_version) not in SpecifierSet(require_version):
                raise VersionIncompatibleError(
                    "Current sceptre version ({0}) does not meet version "
                    "requirements: {1}".format(
                        sceptre_version, require_version
                    )
                )

    def add_resolver_constructors(
            self, environment_config, connection_manager
    ):
        """
        Adds PyYAML constructors for all resolver classes found in the
        resolvers folder within Sceptre library and the current Sceptre project
        folder.

        :param environment_config: A environment config.
        :type environment_config: sceptre.config.Config
        :param connection_manager: A connection manager.
        :type connection_manager: sceptre.connection_manager.ConnectionManager
        """
        self.logger.debug("Adding resolver yaml constructors")

        def resolver_constructor_factory(node_class):
            """
            Returns a lambda function that will contruct objects from a
            given node class.

            :param node_class: A resolver class to construct of objects from.
            :type node_class: class
            :returns: A lambda that constructs resolver objects.
            :rtype: func
            """
            return lambda loader, node: node_class(
                loader.construct_scalar(node),
                connection_manager,
                environment_config,
                self
            )  # pragma: no cover

        resolvers_folder = os.path.join(os.path.dirname(__file__), "resolvers")
        self.add_yaml_constructors(
            resolvers_folder, Resolver, resolver_constructor_factory
        )
        external_resolver_folder = os.path.join(self.sceptre_dir, "resolvers")
        self.add_yaml_constructors(
            external_resolver_folder, Resolver, resolver_constructor_factory
        )

    def add_hook_constructors(self, environment_config, connection_manager):
        """
        Adds PyYAML constructors for all hook classes found in the hooks folder
        within Sceptre library and the current Sceptre project folder.

        :param environment_config: A environment config.
        :type environment_config: Config
        :param connection_manager: A connection manager.
        :type connection_manager: ConnectionManager
        """
        self.logger.debug("Adding hook yaml constructors")

        def hook_constructor_factory(node_class):
            """
            This returns a lambda function that will contruct objects from a
            given node class.

            :param node_class: A hook class to construct of objects from.
            :type node_class: class
            :returns: A lambda that constructs hook objects.
            :rtype: func
            """
            return lambda loader, node: node_class(
                loader.construct_scalar(node),
                connection_manager,
                environment_config,
                self
            )  # pragma: no cover

        library_hook_folder = os.path.join(os.path.dirname(__file__), "hooks")
        self.add_yaml_constructors(
            library_hook_folder, Hook, hook_constructor_factory
        )

        project_hook_folder = os.path.join(self.sceptre_dir, "hooks")
        self.add_yaml_constructors(
            project_hook_folder, Hook, hook_constructor_factory
        )

    def add_yaml_constructors(self, base_directory, base_type, factory):
        """
        Adds PyYAML constructors for all classes which inherit from a
        specific base type within a given directory.

        :param base_directory: A path of a directory to search for classes.
        :type base_directory: str
        :param base_type: The base class in which the class must inherit from.
        :type base_type: class
        :param factory: A function to use to construct objects.
        :type factory: function
        """
        classes = get_subclasses(
            directory=base_directory, class_type=base_type
        )

        for node_name, node_class in classes.items():
            node_tag = u'!' + node_name
            yaml.SafeLoader.add_constructor(
                node_tag, factory(node_class)
            )
            self.logger.debug(
                "Added constructor for %s with node tag %s",
                str(node_class), node_tag
            )



# -*- coding: utf-8 -*-
from colorama import Fore, Style
from mock import Mock,  patch, MagicMock
import pytest

from sceptre.config import Config
from sceptre.hooks.asg_scheduled_actions import ASGScheduledActions
from sceptre.exceptions import InvalidHookArgumentValueError
from sceptre.exceptions import InvalidHookArgumentTypeError


class TestASGScheduledActions(object):
    def setup_method(self, test_method):
        self.deprecation_notice = (
            "{0}The asg_scheduled_actions hook has been "
            "deprecated and will be removed in a later version of Sceptre. "
            "Use the asg_scaling_processes hook instead. Example: "
            "!asg_scaling_processes <suspend|resume>::ScheduledActions{1}"
        ).format(Fore.YELLOW, Style.RESET_ALL)

        self.mock_asg_scheduled_actions = ASGScheduledActions()

    def test_get_stack_resources_sends_correct_request(self):
        mock_environment_config = MagicMock(spec=Config)
        mock_environment_config.__getitem__.return_value = "project_code"
        mock_environment_config.environment_path = "path"
        self.mock_asg_scheduled_actions.environment_config =\
            mock_environment_config
        mock_stack_config = MagicMock(spec=Config)
        mock_stack_config.name = "stack"
        self.mock_asg_scheduled_actions.stack_config = mock_stack_config
        mock_connection_manager = Mock()
        mock_connection_manager.call.return_value = {
            "StackResources": [
                {
                    "ResourceType": "AWS::AutoScaling::AutoScalingGroup",
                    'PhysicalResourceId': 'cloudreach-examples-asg'
                }
            ]
        }
        self.mock_asg_scheduled_actions.\
            connection_manager = mock_connection_manager
        self.mock_asg_scheduled_actions\
            ._get_stack_resources()
        mock_connection_manager.call.\
            assert_called_with(
                service="cloudformation",
                command="describe_stack_resources",
                kwargs={
                    "StackName": "project_code-path-stack",
                }
            )

    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._get_stack_resources"
    )
    def test_find_autoscaling_groups_with_stack_with_asgs(
        self, mock_get_stack_resources
    ):
        mock_get_stack_resources.return_value = [{
            'LogicalResourceId': 'AutoScalingGroup',
            'PhysicalResourceId': 'cloudreach-examples-asg',
            'ResourceStatus': 'CREATE_COMPLETE',
            'ResourceType': 'AWS::AutoScaling::AutoScalingGroup',
            'StackId': 'arn:aws:...',
            'StackName': 'cloudreach-examples-dev-vpc'
        }]
        response = self.mock_asg_scheduled_actions.\
            _find_autoscaling_groups()

        assert response == ["cloudreach-examples-asg"]

    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._get_stack_resources"
    )
    def test_find_autoscaling_groups_with_stack_without_asgs(
        self, mock_get_stack_resources
    ):
        mock_get_stack_resources.return_value = []
        response = self.mock_asg_scheduled_actions.\
            _find_autoscaling_groups()

        assert response == []

    @patch('warnings.warn')
    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._find_autoscaling_groups"
    )
    def test_run_with_resume_argument(
        self, mock_find_autoscaling_groups, mock_warn
    ):
        self.mock_asg_scheduled_actions.argument = u"resume"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.mock_asg_scheduled_actions.connection_manager = Mock()
        self.mock_asg_scheduled_actions.run()
        self.mock_asg_scheduled_actions.connection_manager\
            .call.assert_called_once_with(
                service="autoscaling",
                command="resume_processes",
                kwargs={
                    "AutoScalingGroupName": "autoscaling_group_1",
                    "ScalingProcesses": [
                        "ScheduledActions"
                    ]
                }
            )
        mock_warn.assert_called_once_with(
            self.deprecation_notice,
            DeprecationWarning
        )

    @patch('warnings.warn')
    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._find_autoscaling_groups"
    )
    def test_run_with_suspend_argument(
        self, mock_find_autoscaling_groups, mock_warn
    ):
        self.mock_asg_scheduled_actions.argument = u"suspend"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.mock_asg_scheduled_actions.connection_manager = Mock()
        self.mock_asg_scheduled_actions.run()
        self.mock_asg_scheduled_actions.connection_manager\
            .call.assert_called_once_with(
                service="autoscaling",
                command="suspend_processes",
                kwargs={
                    "AutoScalingGroupName": "autoscaling_group_1",
                    "ScalingProcesses": [
                        "ScheduledActions"
                    ]
                }
            )
        mock_warn.assert_called_once_with(
            self.deprecation_notice,
            DeprecationWarning
        )

    @patch('warnings.warn')
    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._find_autoscaling_groups"
    )
    def test_run_with_invalid_string_argument(
        self, mock_find_autoscaling_groups, mock_warn
    ):
        self.mock_asg_scheduled_actions.argument = u"invalid_string"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.mock_asg_scheduled_actions.connection_manager = Mock()
        with pytest.raises(InvalidHookArgumentValueError):
            self.mock_asg_scheduled_actions.run()
        mock_warn.assert_called_once_with(
            self.deprecation_notice,
            DeprecationWarning
        )

    @patch('warnings.warn')
    @patch(
        "sceptre.hooks.asg_scheduled_actions"
        ".ASGScheduledActions._find_autoscaling_groups"
    )
    def test_run_with_non_string_argument(
        self, mock_find_autoscaling_groups, mock_warn
    ):
        self.mock_asg_scheduled_actions.argument = 10
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.mock_asg_scheduled_actions.connection_manager = Mock()
        with pytest.raises(InvalidHookArgumentTypeError):
            self.mock_asg_scheduled_actions.run()
        mock_warn.assert_called_once_with(
            self.deprecation_notice,
            DeprecationWarning
        )



# -*- coding: utf-8 -*-

import pytest
from mock import patch, sentinel, Mock, MagicMock

import datetime
from dateutil.tz import tzutc

from botocore.exceptions import ClientError

from sceptre.config import Config
from sceptre.stack import Stack
from sceptre.template import Template
from sceptre.stack_status import StackStatus
from sceptre.stack_status import StackChangeSetStatus
from sceptre.exceptions import CannotUpdateFailedStackError
from sceptre.exceptions import UnknownStackStatusError
from sceptre.exceptions import UnknownStackChangeSetStatusError
from sceptre.exceptions import StackDoesNotExistError
from sceptre.exceptions import ProtectedStackError


class TestStack(object):

    @patch("sceptre.stack.Stack.config")
    def setup_method(self, test_method, mock_config):
        self.mock_environment_config = MagicMock(spec=Config)
        self.mock_environment_config.environment_path = sentinel.path
        # environment config is an object which inherits from dict. Its
        # attributes are accessable via dot and square bracket notation.
        # In order to mimic the behaviour of the square bracket notation,
        # a side effect is used to return the expected value from the call to
        # __getitem__ that the square bracket notation makes.
        self.mock_environment_config.__getitem__.side_effect = [
            sentinel.project_code,
            sentinel.region
        ]
        self.mock_connection_manager = Mock()

        self.stack = Stack(
            name="stack_name",
            environment_config=self.mock_environment_config,
            connection_manager=self.mock_connection_manager
        )

        # Set default value for stack properties
        self.stack._external_name = sentinel.external_name

    def test_initiate_stack(self):
        assert self.stack.name == "stack_name"
        assert self.stack.environment_config == self.mock_environment_config
        assert self.stack.project == sentinel.project_code
        assert self.stack._environment_path == sentinel.path
        assert self.stack._config is None
        assert self.stack._template is None
        assert self.stack.region == sentinel.region
        assert self.stack.connection_manager == self.mock_connection_manager
        assert self.stack._hooks is None
        assert self.stack._dependencies is None

    @patch("sceptre.stack.Stack.config")
    def test_initialiser_calls_correct_methods(self, mock_config):
        mock_config.get.return_value = sentinel.hooks
        self.stack._config = {
            "parameters": sentinel.parameters,
            "hooks": sentinel.hooks
        }
        self.mock_environment_config = MagicMock(spec=Config)
        self.mock_environment_config.environment_path = sentinel.path
        # environment config is an object which inherits from dict. Its
        # attributes are accessable via dot and square bracket notation.
        # In order to mimic the behaviour of the square bracket notation,
        # a side effect is used to return the expected value from the call to
        # __getitem__ that the square bracket notation makes.
        self.mock_environment_config.__getitem__.side_effect = [
            sentinel.project_code,
            sentinel.template_bucket_name,
            sentinel.region
        ]

        Stack(
            name=sentinel.name,
            environment_config=self.mock_environment_config,
            connection_manager=sentinel.connection_manager
        )

    def test_repr(self):
        self.stack.name = "stack_name"
        self.stack.environment_config = {"key": "val"}
        self.stack.connection_manager = "connection_manager"
        assert self.stack.__repr__() == \
            "sceptre.stack.Stack(stack_name='stack_name', \
environment_config={'key': 'val'}, connection_manager=connection_manager)"

    @patch("sceptre.stack.Config")
    def test_config_loads_config(self, mock_Config):
        self.stack._config = None
        self.stack.name = "stack"
        # self.stack.environment_config = MagicMock(spec=Config)
        self.stack.environment_config.sceptre_dir = sentinel.sceptre_dir
        self.stack.environment_config.environment_path = \
            sentinel.environment_path
        self.stack.environment_config.get.return_value = \
            sentinel.user_variables
        mock_config = Mock()
        mock_Config.with_yaml_constructors.return_value = mock_config

        response = self.stack.config
        mock_Config.with_yaml_constructors.assert_called_once_with(
            sceptre_dir=sentinel.sceptre_dir,
            environment_path=sentinel.environment_path,
            base_file_name="stack",
            environment_config=self.stack.environment_config,
            connection_manager=self.stack.connection_manager
        )
        mock_config.read.assert_called_once_with(sentinel.user_variables,
                                                 self.stack.environment_config)
        assert response == mock_config

    def test_config_returns_config_if_it_exists(self):
        self.stack._config = sentinel.config
        response = self.stack.config
        assert response == sentinel.config

    def test_dependencies_loads_dependencies(self):
        self.stack.name = "dev/security-group"
        self.stack._config = {
            "dependencies": ["dev/vpc", "dev/vpc", "dev/subnets"]
        }
        dependencies = self.stack.dependencies
        assert dependencies == set(["dev/vpc", "dev/subnets"])

    def test_dependencies_returns_dependencies_if_it_exists(self):
        self.stack._dependencies = sentinel.dependencies
        response = self.stack.dependencies
        assert response == sentinel.dependencies

    def test_hooks_with_no_cache(self):
        self.stack._hooks = None
        self.stack._config = {}
        self.stack._config["hooks"] = sentinel.hooks

        assert self.stack.hooks == sentinel.hooks

    def test_hooks_with_cache(self):
        self.stack._hooks = sentinel.hooks
        assert self.stack.hooks == sentinel.hooks

    @patch("sceptre.stack.Template")
    def test_template_loads_template(self, mock_Template):
        self.stack._template = None
        self.stack.environment_config.sceptre_dir = "sceptre_dir"
        self.stack._config = {
            "template_path": "template_path",
            "sceptre_user_data": sentinel.sceptre_user_data
        }
        mock_Template.return_value = sentinel.template

        response = self.stack.template

        mock_Template.assert_called_once_with(
            path="sceptre_dir/template_path",
            sceptre_user_data=sentinel.sceptre_user_data
        )
        assert response == sentinel.template

    def test_template_returns_template_if_it_exists(self):
        self.stack._template = sentinel.template
        response = self.stack.template
        assert response == sentinel.template

    @patch("sceptre.stack.get_external_stack_name")
    def test_external_name_with_custom_stack_name(
            self, mock_get_external_stack_name
    ):
        self.stack._external_name = None

        self.stack._config = {"stack_name": "custom_stack_name"}
        external_name = self.stack.external_name
        assert external_name == "custom_stack_name"

    def test_external_name_without_custom_name(self):
        self.stack._external_name = None
        self.stack.project = "project"
        self.stack.name = "stack-name"
        self.stack._config = {}

        external_name = self.stack.external_name
        assert external_name == "project-stack-name"

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_create_sends_correct_request(
        self, mock_get_template_details,
        mock_wait_for_completion, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {"stack_tags": {
            "tag1": "val1"
        }}
        self.stack._hooks = {}
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.config["notifications"] = [sentinel.notification]
        self.stack.create()

        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_create_sends_correct_request_no_notifications(
            self, mock_get_template_details,
            mock_wait_for_completion, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {"stack_tags": {
            "tag1": "val1"
        }}
        self.stack._hooks = {}
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.create()

        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_create_sends_correct_request_with_failure(
        self, mock_get_template_details,
        mock_wait_for_completion, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {"stack_tags": {
            "tag1": "val1"
        }}
        self.stack._hooks = {}
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.config["notifications"] = [sentinel.notification]
        self.stack.config["on_failure"] = 'DO_NOTHING'
        self.stack.create()

        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ],
                "OnFailure": 'DO_NOTHING'
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_update_sends_correct_request(
        self, mock_get_template_details,
        mock_wait_for_completion, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {"stack_tags": {
            "tag1": "val1"
        }}
        self.stack._hooks = {}
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.config["notifications"] = [sentinel.notification]

        self.stack.update()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_update_sends_correct_request_no_notification(
            self, mock_get_template_details,
            mock_wait_for_completion, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {"stack_tags": {
            "tag1": "val1"
        }}
        self.stack._hooks = {}
        self.stack.config["role_arn"] = sentinel.role_arn

        self.stack.update()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.create")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_stack_that_does_not_exist(
            self, mock_get_status, mock_create, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.side_effect = StackDoesNotExistError()
        mock_create.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.create")
    @patch("sceptre.stack.Stack.delete")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_stack_that_failed_to_create(
            self, mock_get_status, mock_delete, mock_create, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_FAILED"
        mock_create.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_delete.assert_called_once_with()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.update")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_updates_to_perform(
            self, mock_get_status, mock_update, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_update.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.update")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_no_updates_to_perform(
            self, mock_get_status, mock_update, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.side_effect = ClientError(
            {
                "Error": {
                    "Code": "NoUpdateToPerformError",
                    "Message": "No updates are to be performed."
                }
            },
            sentinel.operation
        )
        response = self.stack.launch()
        mock_update.assert_called_once_with()
        assert response == StackStatus.COMPLETE

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.update")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_unknown_client_error(
            self, mock_get_status, mock_update, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.side_effect = ClientError(
            {
                "Error": {
                    "Code": "Boom!",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.launch()

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_in_progress_stack(self, mock_get_status, mock_hooks):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_IN_PROGRESS"
        response = self.stack.launch()
        assert response == StackStatus.IN_PROGRESS

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_failed_stack(self, mock_get_status, mock_hooks):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "UPDATE_FAILED"
        with pytest.raises(CannotUpdateFailedStackError):
            response = self.stack.launch()
            assert response == StackStatus.FAILED

    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_launch_with_unknown_stack_status(
            self, mock_get_status, mock_hooks
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "UNKNOWN_STATUS"
        with pytest.raises(UnknownStackStatusError):
            self.stack.launch()

    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_delete_with_created_stack(
            self, mock_get_status, mock_hooks, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.delete()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_stack_does_not_exist_error(
            self, mock_get_status, mock_hooks, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        self.stack.config["role_arn"] = sentinel.role_arn
        mock_wait_for_completion.side_effect = StackDoesNotExistError()
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_non_existent_client_error(
            self, mock_get_status, mock_hooks, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        self.stack.config["role_arn"] = sentinel.role_arn
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_unexpected_client_error(
            self, mock_get_status, mock_hooks, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        mock_get_status.return_value = "CREATE_COMPLETE"
        self.stack.config["role_arn"] = sentinel.role_arn
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.delete()

    @patch("sceptre.stack.Stack._wait_for_completion")
    @patch("sceptre.stack.Stack.hooks")
    @patch("sceptre.stack.Stack.get_status")
    def test_delete_with_non_existent_stack(
            self, mock_get_status, mock_hooks, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        mock_get_status.side_effect = StackDoesNotExistError()
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    def test_describe_stack_sends_correct_request(self):
        self.stack.describe()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stacks",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_events_sends_correct_request(self):
        self.stack.describe_events()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_events",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_resources_sends_correct_request(self):
        self.stack.connection_manager.call.return_value = {
            "StackResources": [
                {
                    "LogicalResourceId": sentinel.logical_resource_id,
                    "PhysicalResourceId": sentinel.physical_resource_id,
                    "OtherParam": sentinel.other_param
                }
            ]
        }
        response = self.stack.describe_resources()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_resources",
            kwargs={"StackName": sentinel.external_name}
        )
        assert response == [
            {
                "LogicalResourceId": sentinel.logical_resource_id,
                "PhysicalResourceId": sentinel.physical_resource_id
            }
        ]

    @patch("sceptre.stack.Stack.describe")
    def test_describe_outputs_sends_correct_request(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{
                "Outputs": sentinel.outputs
            }]
        }
        response = self.stack.describe_outputs()
        mock_describe.assert_called_once_with()
        assert response == sentinel.outputs

    @patch("sceptre.stack.Stack.describe")
    def test_describe_outputs_handles_stack_with_no_outputs(
            self, mock_describe
    ):
        mock_describe.return_value = {
            "Stacks": [{}]
        }
        response = self.stack.describe_outputs()
        assert response == []

    def test_continue_update_rollback_sends_correct_request(self):
        self.stack._config = {
            "template_path": sentinel.template_path,
        }
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.continue_update_rollback()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="continue_update_rollback",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    def test_set_stack_policy_sends_correct_request(self):
        self.stack.set_policy("tests/fixtures/stack_policies/unlock.json")
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="set_stack_policy",
            kwargs={
                "StackName": sentinel.external_name,
                "StackPolicyBody": """{
  "Statement" : [
    {
      "Effect" : "Allow",
      "Action" : "Update:*",
      "Principal": "*",
      "Resource" : "*"
    }
  ]
}
"""
            }
        )

    def test_get_stack_policy_sends_correct_request(self):
        self.stack.get_policy()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="get_stack_policy",
            kwargs={
                "StackName": sentinel.external_name
            }
        )

    @patch("sceptre.stack.Stack._get_template_details")
    def test_validate_template_sends_correct_request(
        self, mock_get_template_details
    ):
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name
        }
        self.stack.validate_template()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="validate_template",
            kwargs={"Template": sentinel.template}
        )

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_create_change_set_sends_correct_request(
        self, mock_get_template_details, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {
            "stack_tags": {"tag1": "val1"}
        }
        self.stack.config["role_arn"] = sentinel.role_arn
        self.stack.config["notifications"] = [sentinel.notification]

        self.stack.create_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    @patch("sceptre.stack.Stack._format_parameters")
    @patch("sceptre.stack.Stack._get_template_details")
    def test_create_change_set_sends_correct_request_no_notifications(
            self, mock_get_template_details, mock_format_params
    ):
        mock_format_params.return_value = sentinel.parameters
        mock_get_template_details.return_value = {
            "Template": sentinel.template
        }
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }
        self.stack._config = {
            "stack_tags": {"tag1": "val1"}
        }
        self.stack.config["role_arn"] = sentinel.role_arn

        self.stack.create_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": sentinel.parameters,
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    def test_delete_change_set_sends_correct_request(self):
        self.stack.delete_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    def test_describe_change_set_sends_correct_request(self):
        self.stack.describe_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    @patch("sceptre.stack.Stack._wait_for_completion")
    def test_execute_change_set_sends_correct_request(
        self, mock_wait_for_completion
    ):
        self.stack._config = {"protect": False}
        self.stack.execute_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="execute_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    def test_list_change_sets_sends_correct_request(self):
        self.stack.list_change_sets()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="list_change_sets",
            kwargs={"StackName": sentinel.external_name}
        )

    @patch("sceptre.stack.Stack.set_policy")
    @patch("os.path.join")
    def test_lock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/lock.json"
        self.stack.lock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/lock.json"
        )

    @patch("sceptre.stack.Stack.set_policy")
    @patch("os.path.join")
    def test_unlock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/unlock.json"
        self.stack.unlock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/unlock.json"
        )

    def test_format_parameters_with_sting_values(self):
        parameters = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3"
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key2", "ParameterValue": "value2"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_none_values(self):
        parameters = {
            "key1": None,
            "key2": None,
            "key3": None
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == []

    def test_format_parameters_with_none_and_string_values(self):
        parameters = {
            "key1": "value1",
            "key2": None,
            "key3": "value3"
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": ["value4", "value5", "value6"],
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4,value5,value6"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_none_and_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": None,
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": ["value5", "value6", "value7"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
            {"ParameterKey": "key3", "ParameterValue": "value5,value6,value7"}
        ]

    def test_format_parameters_with_none_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": None
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
        ]

    @patch("sceptre.stack.Stack.describe")
    def test_get_status_with_created_stack(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{"StackStatus": "CREATE_COMPLETE"}]
        }
        status = self.stack.get_status()
        assert status == "CREATE_COMPLETE"

    @patch("sceptre.stack.Stack.describe")
    def test_get_status_with_non_existent_stack(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        with pytest.raises(StackDoesNotExistError):
            self.stack.get_status()

    @patch("sceptre.stack.Stack.describe")
    def test_get_status_with_unknown_clinet_error(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.get_status()

    def test_get_template_details_with_upload(self):
        self.stack._template = Mock(spec=Template)
        self.stack._template.upload_to_s3.return_value = sentinel.template_url
        self.stack.environment_config = {
            "template_bucket_name": sentinel.template_bucket_name,
            "template_key_prefix": sentinel.template_key_prefix
        }

        template_details = self.stack._get_template_details()

        self.stack._template.upload_to_s3.assert_called_once_with(
            self.stack.region,
            sentinel.template_bucket_name,
            sentinel.template_key_prefix,
            self.stack._environment_path,
            sentinel.external_name,
            self.stack.connection_manager
        )

        assert template_details == {"TemplateURL": sentinel.template_url}

    def test_get_template_details_without_upload(self):
        self.stack._template = Mock(spec=Template)
        self.stack._template.body = sentinel.body
        self.stack.environment_config = {
            "template_key_prefix": sentinel.template_key_prefix
        }

        template_details = self.stack._get_template_details()

        assert template_details == {"TemplateBody": sentinel.body}

    def test_get_role_arn_without_role(self):
        self.stack._template = Mock(spec=Template)
        self.stack._config = {
            "template_path": sentinel.template_path,
        }
        assert self.stack._get_role_arn() == {}

    def test_get_role_arn_with_role(self):
        self.stack._template = Mock(spec=Template)
        self.stack._config = {
            "template_path": sentinel.template_path,
        }
        self.stack.config["role_arn"] = sentinel.role_arn
        assert self.stack._get_role_arn() == {"RoleARN": sentinel.role_arn}

    def test_protect_execution_without_protection(self):
        self.stack._config = {"protect": False}
        # Function should do nothing if protect == False
        self.stack._protect_execution()

    def test_protect_execution_without_explicit_protection(self):
        self.stack._config = {}
        # Function should do nothing if protect isn't explicitly set
        self.stack._protect_execution()

    def test_protect_execution_with_protection(self):
        self.stack._config = {"protect": True}
        with pytest.raises(ProtectedStackError):
            self.stack._protect_execution()

    @patch("sceptre.stack.time")
    @patch("sceptre.stack.Stack._log_new_events")
    @patch("sceptre.stack.Stack.get_status")
    @patch("sceptre.stack.Stack._get_simplified_status")
    def test_wait_for_completion_calls_log_new_events(
            self, mock_get_simplified_status, mock_get_status,
            mock_log_new_events, mock_time
    ):
        mock_get_simplified_status.return_value = StackStatus.COMPLETE

        self.stack._wait_for_completion()
        mock_log_new_events.assert_called_once_with()

    @pytest.mark.parametrize("test_input,expected", [
        ("ROLLBACK_COMPLETE", StackStatus.FAILED),
        ("STACK_COMPLETE", StackStatus.COMPLETE),
        ("STACK_IN_PROGRESS", StackStatus.IN_PROGRESS),
        ("STACK_FAILED", StackStatus.FAILED)
    ])
    def test_get_simplified_status_with_known_stack_statuses(
            self, test_input, expected
    ):
        response = self.stack._get_simplified_status(test_input)
        assert response == expected

    def test_get_simplified_status_with_stack_in_unknown_state(self):
        with pytest.raises(UnknownStackStatusError):
            self.stack._get_simplified_status("UNKOWN_STATUS")

    @patch("sceptre.stack.Stack.describe_events")
    def test_log_new_events_calls_describe_events(self, mock_describe_events):
        mock_describe_events.return_value = {
            "StackEvents": []
        }
        self.stack._log_new_events()
        self.stack.describe_events.assert_called_once_with()

    @patch("sceptre.stack.Stack.describe_events")
    def test_log_new_events_prints_correct_event(self, mock_describe_events):
        mock_describe_events.return_value = {
            "StackEvents": [
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 2, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-2",
                    "ResourceType": "type-2",
                    "ResourceStatus": "resource-status"
                },
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 1, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-1",
                    "ResourceType": "type-1",
                    "ResourceStatus": "resource",
                    "ResourceStatusReason": "User Initiated"
                }
            ]
        }
        self.stack.most_recent_event_datetime = (
            datetime.datetime(2016, 3, 15, 14, 0, 0, 0, tzinfo=tzutc())
        )
        self.stack._log_new_events()

    @patch("sceptre.stack.time")
    @patch("sceptre.stack.Stack._get_cs_status")
    def test_wait_for_cs_completion_calls_get_cs_status(
        self, mock_get_cs_status, mock_time
    ):
        mock_get_cs_status.side_effect = [
            StackChangeSetStatus.PENDING, StackChangeSetStatus.READY
        ]

        self.stack.wait_for_cs_completion(sentinel.change_set_name)
        mock_get_cs_status.assert_called_with(sentinel.change_set_name)

    @patch("sceptre.stack.Stack.describe_change_set")
    def test_get_cs_status_handles_all_statuses(
        self, mock_describe_change_set
    ):
        scss = StackChangeSetStatus
        return_values = {                                                                                                     # NOQA
                 "Status":    ('CREATE_PENDING', 'CREATE_IN_PROGRESS', 'CREATE_COMPLETE', 'DELETE_COMPLETE', 'FAILED'),       # NOQA
        "ExecutionStatus": {                                                                                                  # NOQA
        'UNAVAILABLE':         (scss.PENDING,     scss.PENDING,         scss.PENDING,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'AVAILABLE':           (scss.PENDING,     scss.PENDING,         scss.READY,        scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_IN_PROGRESS': (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_COMPLETE':    (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_FAILED':      (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'OBSOLETE':            (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        }                                                                                                                     # NOQA
        }                                                                                                                     # NOQA

        for i, status in enumerate(return_values['Status']):
            for exec_status, returns in \
                    return_values['ExecutionStatus'].items():
                mock_describe_change_set.return_value = {
                    "Status": status,
                    "ExecutionStatus": exec_status
                }
                response = self.stack._get_cs_status(sentinel.change_set_name)
                assert response == returns[i]

        for status in return_values['Status']:
            mock_describe_change_set.return_value = {
                "Status": status,
                "ExecutionStatus": 'UNKOWN_STATUS'
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.stack._get_cs_status(sentinel.change_set_name)

        for exec_status in return_values['ExecutionStatus'].keys():
            mock_describe_change_set.return_value = {
                "Status": 'UNKOWN_STATUS',
                "ExecutionStatus": exec_status
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.stack._get_cs_status(sentinel.change_set_name)

        mock_describe_change_set.return_value = {
            "Status": 'UNKOWN_STATUS',
            "ExecutionStatus": 'UNKOWN_STATUS',
        }
        with pytest.raises(UnknownStackChangeSetStatusError):
            self.stack._get_cs_status(sentinel.change_set_name)

    @patch("sceptre.stack.Stack.describe_change_set")
    def test_get_cs_status_raises_unexpected_exceptions(
        self, mock_describe_change_set
    ):
        mock_describe_change_set.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ChangeSetNotFound",
                    "Message": "ChangeSet [*] does not exist"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack._get_cs_status(sentinel.change_set_name)



import abc
import logging
from functools import wraps


class Hook(object):
    """
    Hook is an abstract base class that should be inherited by all
    hooks used in hooks. Environment and stack config and the connection
    manager are supplied to the class, as they may be of use to inheriting
    classes.

    :param argument: The argument in which the task can use.
    :type argument: str
    :param environment_config: The environment_config from config.yaml files.
    :type environment_config: sceptre.config.Config
    :param stack_config: The stack config of the stack that this parameter is \
    associated with.
    :type stack_config: sceptre.config.Config
    :param connection_manager: A connection_manager.
    :type connection_manager: sceptre.connection_manager.ConnectionManager
    """
    __metaclass__ = abc.ABCMeta

    def __init__(
            self, argument=None, connection_manager=None,
            environment_config=None, stack_config=None):
        self.logger = logging.getLogger(__name__)
        self.argument = argument
        self.connection_manager = connection_manager
        self.environment_config = environment_config
        self.stack_config = stack_config

    @abc.abstractmethod
    def run(self):
        """
        run is an abstract method which must be overwritten by all
        inheriting classes. Run should execute the logic of the hook.
        """
        pass  # pragma: no cover


def execute_hooks(hooks):
    """
    Searches through dictionary or list for Resolver objects and replaces
    them with the resolved value. Supports nested dictionaries and lists.
    Does not detect Resolver objects used as keys in dictionaries.

    :param attr: A complex data structure to search through.
    :type attr: dict or list
    :return: A complex data structure without Resolver objects.
    :rtype: dict or list
    """
    if isinstance(hooks, list):
        for hook in hooks:
            if isinstance(hook, Hook):
                hook.run()


def add_stack_hooks(func):
    """
    A function decorator to trigger the before and after hooks, relative
    to the decorated function's name.
    :param func: a function that operates on a stack
    :type func: function
    """
    @wraps(func)
    def decorated(self, *args, **kwargs):
        execute_hooks(self.hooks.get("before_" + func.__name__))
        response = func(self, *args, **kwargs)
        execute_hooks(self.hooks.get("after_" + func.__name__))

        return response

    return decorated



# -*- coding: utf-8 -*-
import warnings
from six import string_types

from colorama import Fore, Style

from sceptre.hooks import Hook
from sceptre.exceptions import InvalidHookArgumentValueError
from sceptre.exceptions import InvalidHookArgumentTypeError


class ASGScheduledActions(Hook):
    """
    This hook has been deprecated in favor of the asg_scaling_processes hook.

    A command to resume or suspend autoscaling group scheduled actions. This is
    useful as schedule actions must be suspended when updating stacks with
    on autoscaling groups.
    """

    def __init__(self, *args, **kwargs):
        super(ASGScheduledActions, self).__init__(*args, **kwargs)

    def run(self):
        """
        Either suspends or resumes any scheduled actions on all autoscaling
        groups with in the current stack.
        """
        warnings.warn(
            "{0}The asg_scheduled_actions hook has been deprecated and will "
            "be removed in a later version of Sceptre. Use the "
            "asg_scaling_processes hook instead. Example: "
            "!asg_scaling_processes <suspend|resume>::ScheduledActions{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )
        if not isinstance(self.argument, string_types):
            raise InvalidHookArgumentTypeError(
                'The argument "{0}" is the wrong type - asg_scheduled_actions '
                'hooks require arguments of type string.'.format(self.argument)
            )
        if self.argument not in ["resume", "suspend"]:
            raise InvalidHookArgumentValueError(
                'The argument "{0}" is invalid - valid arguments for '
                'asg_scheduled_actions hooks are "resume" or "suspend".'
                .format(self.argument)
            )

        self.argument += "_processes"

        autoscaling_group_names = self._find_autoscaling_groups()
        for autoscaling_group in autoscaling_group_names:
            self.connection_manager.call(
                service="autoscaling",
                command=self.argument,
                kwargs={
                    "AutoScalingGroupName": autoscaling_group,
                    "ScalingProcesses": [
                        "ScheduledActions"
                    ]
                }
            )

    def _get_stack_resources(self):
        """
        Retrieves all resources in stack.
        :return: list
        """
        full_stack_name = "-".join([
            self.environment_config["project_code"],
            self.environment_config.environment_path,
            self.stack_config.name
        ]).replace("/", "-")
        response = self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_resources",
            kwargs={"StackName": full_stack_name}
        )
        return response.get("StackResources", [])

    def _find_autoscaling_groups(self):
        """
        Retrieves all the autoscaling groups
        :return: list [str]
        """
        asg_names = []
        resources = self._get_stack_resources()
        resource_type = "AWS::AutoScaling::AutoScalingGroup"
        for resource in resources:
            if resource.get("ResourceType", False) == resource_type:
                    asg_names.append(resource["PhysicalResourceId"])

        return asg_names



# -*- coding: utf-8 -*-

import os
import yaml
import warnings

from colorama import Fore, Style

from sceptre.resolvers import Resolver


class ProjectVariables(Resolver):
    """
    Resolver for the configuration in a file.

    :param argument: Configuration value to be returned.
    :type argument: str
    """

    def __init__(self, *args, **kwargs):
        super(ProjectVariables, self).__init__(*args, **kwargs)

    def resolve(self):
        """
        Searches for value at given key location.

        :returns: Value at argument location.
        :rtype: str
        """
        warnings.warn(
            "{0}The project_variables resolver has been deprecated, and will "
            "be removed in a later version of Sceptre. Depending on your "
            "use case, you may find user variables appropiate. "
            "Example: sceptre --var \"iam_role=<your iam role>\" <COMMAND>{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )
        file_path = os.path.join(
            self.environment_config.sceptre_dir,
            self.argument
        )

        with open(file_path, "r") as f:
            variables_dict = yaml.safe_load(f)

        dict_keys = self.environment_config.environment_path.split("/")
        dict_keys.append(self.stack_config.name)

        for key in dict_keys:
            variables_dict = variables_dict[key]

        return variables_dict



import subprocess
import warnings
from six import string_types

from colorama import Fore, Style

from sceptre.hooks import Hook
from sceptre.exceptions import InvalidHookArgumentTypeError


class Bash(Hook):
    """
    This hook has been deprecated in favor of cmd hook.
    This command with execute the argument string with bash.

    """
    ALLOW_COMMAND_ERROR = True

    def __init__(self, *args, **kwargs):
        super(Bash, self).__init__(*args, **kwargs)

    def run(self):
        """
        Runs argument string in child process with bash.

        :raise: sceptre.exceptions.InvalidTaskArgumentTypeException
        """
        warnings.warn(
            "{0}The bash hook has been deprecated, and will "
            "be removed in a later version of Sceptre. "
            "Use the cmd hook instead. Example: !cmd <command>{1}"
            .format(Fore.YELLOW, Style.RESET_ALL),
            DeprecationWarning
        )
        if not isinstance(self.argument, string_types):
            raise InvalidHookArgumentTypeError(
                'The argument "{0}" is the wrong type - Bash hooks require '
                'arguments of type string.'.format(self.argument)
            )

        if Bash.ALLOW_COMMAND_ERROR:
            subprocess.call(["/bin/bash", "-c", self.argument])
        else:
            subprocess.check_call(["/bin/bash", "-c", self.argument])
